import unittest
try:
    from Scenario2.interpolators import *
    from Scenario2.keyframes import KF, ValueGenerator, Interval
except:
    print "ERROR: failed to import scenario.interpolators"


    
import types
from math import *
import numpy

def sameIntSequences(list1, list2):
    for a,b in zip(list1,list2):
        if a!=b:
            return False
    return True

def sameFloatSequences(list1, list2):
    for a,b in zip(list1,list2):
        try:
            len(a)
            len(b)
            for c,d in zip(a, b):
                if "%.2f"%c !=  "%.2f"%c:
                   print  a , "!=", b
                   return False
        except:
            if "%.2f"%a !=  "%.2f"%b:
                print  "%.2f"%a , "%.2f"%b
                return False
    return True

def adding(x):
    return x+x
     
def withargs(x,y,height,width):
    z=x*y*height*width
    return z
    
def simple(x):
    return x

def sinfunc(mod,x):
    return round(sin(mod))
    
def cosfunc(x):
    return cos(x)


# BooleanInterpolator, CompositeInterpolator,  ReadDataInterpolator

class ValueGeneratorTest(unittest.TestCase):


    def test_IntScalarInterpolators(self):

        kf1 = KF(0, 0)
        kf2 = KF(20, 10)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        ip = IntScalarInterpolator(None, None)
        i1.setValGenGenerator(ip)
        
        val = i1.valGen.getValue(0.0, i1)
        self.assertEqual(val, i1.valGen.generator.firstVal)
        self.assertEqual(val, 0)
        
        val = i1.valGen.getValue(1.0, i1)
        self.assertEqual( val, i1.valGen.generator.lastVal)
        self.assertEqual(val, 10)
        
        self.assertEqual( i1.valGen.getValue(0.3, i1), 3)
        self.assertEqual( i1.valGen.getValue(0.34, i1), 3)
        self.assertEqual( i1.valGen.getValue(0.38, i1), 4)
        
        #test configure()
        i1.valGen.generator.configure(firstVal = 25, lastVal = 30)
        val = i1.valGen.getValue(0.0, i1)
        self.assertEqual(val, 25)
        val = i1.valGen.getValue(1.0, i1)
        self.assertEqual( val, 30)
        self.assertEqual(i1.valGen.generator.valueRange, 5)
        self.assertEqual( i1.valGen.getValue(0.5, i1), 28)
        

    def test_FloatScalarInterpolator(self):

        kf1 = KF(0, 0.)
        kf2 = KF(20, 10.)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        
        ip = FloatScalarInterpolator(None, None)
        bl = BehaviorList(None, None)
        bl.addBehavior(ip)
        i1.setValGenGenerator(bl)
        
        val = i1.valGen.getValue(0.0, i1)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual(val, firstVal)
        self.assertEqual(0., firstVal)
        
        val = i1.valGen.getValue(1.0, i1)
        lastVal = i1.valGen.generator.lastVal
        self.assertEqual( val, lastVal)
        self.assertEqual( 10., lastVal)
        
        val = "%.2f" % i1.valGen.getValue(0.35, i1)
        self.assertEqual(val, "3.50")
        val = "%.2f" % i1.valGen.getValue(0.5, i1)
        self.assertEqual(val, "5.00")
        
        #testconfigure()
        i1.valGen.generator.configure(firstVal = 25.5, lastVal = 30.5)
        val = "%.2f" % i1.valGen.getValue(0.0, i1)
        self.assertEqual(val, "25.50")
        val = "%.2f" % i1.valGen.getValue(1.0, i1)
        self.assertEqual( val, "30.50")
        self.assertEqual("%.2f" % i1.valGen.generator.behaviors[0].valueRange, "5.00")
        val =  "%.2f" % i1.valGen.getValue(0.56, i1)
        self.assertEqual( val, "28.30")
        
        i1.valGen.generator.configure(active = False)
        self.assertEqual(ip.active, False)
    


    def test_IntVarScalarInterpolator(self):
        kf1 = KF(0, 1)
        kf2 = KF(20, [2,4,6])
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        
        bl = BehaviorList(None, None)
        ip = IntVarScalarInterpolator (None, None)
        bl.addBehavior(ip)
        i1.setValGenGenerator(bl)

        val = i1.valGen.getValue(0.0, i1 )
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual(val, [firstVal,])
        self.assertEqual(firstVal, 1)
        
        val = i1.valGen.getValue(1.0, i1)
        lastVal = i1.valGen.generator.lastVal
        self.assertEqual( val, lastVal)
        self.assertEqual(lastVal, [2,4,6])
        
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(val, [2, 3, 4])
        val = i1.valGen.getValue(0.6, i1)
        self.assertEqual(val,[2, 3, 4])
        val = i1.valGen.getValue(0.67, i1)
        self.assertEqual(val, [2, 3, 4])  
        # configure()
        i1.valGen.generator.configure(firstVal = [2,4,6], lastVal = 1)
        self.assertEqual(i1.valGen.generator.firstVal ,[2,4,6])
        self.assertEqual(i1.valGen.generator.lastVal , 1)
        
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(val, [2, 3, 4])
        val = i1.valGen.getValue(0.6, i1)
        self.assertEqual(val, [1, 2, 3])
        val = i1.valGen.getValue(0.67, i1)
        self.assertEqual(val, [1, 2, 3])
        
        self.assertEqual(i1.valGen.generator.nbvar , 3)


    def test_FloatVarScalarInterpolator(self):
        kf1 = KF(0, 1.)
        kf2 = KF(20, [2.,4.,6.])
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        ip = FloatVarScalarInterpolator (None, None)
        i1.setValGenGenerator(ip)

        val = i1.valGen.getValue(0.0, i1)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual(val, firstVal)
        self.assertEqual(firstVal, [1.])
        
        val = i1.valGen.getValue(1.0, i1)
        lastVal = i1.valGen.generator.lastVal
        self.assertEqual( val, lastVal)
        self.assertEqual(sameFloatSequences(lastVal, [2.,4.,6.]), True)
        
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(sameFloatSequences(val, [1.5, 2.5, 3.5]), True)
        val = i1.valGen.getValue(0.6, i1)
        self.assertEqual(sameFloatSequences(val,[1.60, 2.799, 4.0]), True)
        val = i1.valGen.getValue(0.67, i1)
        self.assertEqual(sameFloatSequences(val, [1.6699, 3.010, 4.34999]), True)
        
        # configure()
        i1.valGen.generator.configure(firstVal = [2.,4.,6.], lastVal = 1)
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(sameFloatSequences(val, [1.5, 2.5, 3.5]), True)
        val = i1.valGen.getValue(0.6, i1)
        self.assertEqual(sameFloatSequences(val, [1.399, 2.200, 3.0]), True)
        val = i1.valGen.getValue(0.67, i1)
        self.assertEqual(sameFloatSequences(val,[1.330, 1.9899, 2.6499] ), True)
        self.assertEqual(i1.valGen.generator.nbvar , 3)


    def test_IntVectorInterpolator(self):
        starts = [0, 10,  -5, -15]
        ends =  [10,  0, -15,  -5]
        kf1 = KF(0, starts)
        kf2 = KF(20, ends)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        bl = BehaviorList(None, None)
        ip = IntVectorInterpolator(None, None)
        bl.addBehavior(ip)
        i1.setValGenGenerator(bl)

        val = i1.valGen.getValue(0.0, i1)
        self.assertEqual( len(val)==4,True)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual(val, firstVal)
        self.assertEqual(firstVal, starts)
        
                        
        val = i1.valGen.getValue(1.0, i1)
        lastVal = i1.valGen.generator.lastVal 
        self.assertEqual(val, lastVal)
        self.assertEqual(lastVal, ends)

    
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual( sameIntSequences(val, [5, 5, -10, -10]),True)
        val = i1.valGen.getValue(0.32, i1)
        self.assertEqual( sameIntSequences(val, [3, 7, -8, -12]),True)
    
        val = i1.valGen.getValue(0.38, i1)
        self.assertEqual( sameIntSequences(val, [4, 6, -9, -11]),True)

        # test configure()
        ends = [10, 20, 0, 0]
        i1.valGen.generator.configure(lastVal = numpy.array(ends ))
        self.assertEqual( sameIntSequences(i1.valGen.generator.behaviors[0].valueRange, [10, 10, 5, 15]), True)
        self.assertEqual( sameIntSequences(i1.valGen.getValue(0.5, i1), [5, 15, -3, -8]), True)
        self.assertEqual( sameIntSequences(i1.valGen.getValue(0.32, i1),[3, 13, -3, -10]), True)
        self.assertEqual( sameIntSequences(i1.valGen.getValue(0.38, i1), [4, 14, -3, -9]), True)

        
    def test_FloatVectorInterpolator(self):
        starts = [0.0, 10.0,  -5.0, -15.0]
        ends =  [10.0,  0.0, -15.0,  -5.0]
        kf1 = KF(0, starts)
        kf2 = KF(20, ends)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        ip = FloatVectorInterpolator(None, None)
        i1.setValGenGenerator(ip)
        
        val = i1.valGen.getValue(0.0, i1)
        self.assertEqual( len(val)==4,True)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual( sameFloatSequences(val, firstVal),True)
        self.assertEqual(sameFloatSequences(firstVal, starts),True)
                        
        val = i1.valGen.getValue(1.0, i1)
        lastVal = i1.valGen.generator.lastVal 
        self.assertEqual( sameFloatSequences(val, lastVal),True)
        self.assertEqual(sameFloatSequences(lastVal, ends),True)
    
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual( sameFloatSequences(val, [5.0, 5.0, -10.0, -10.0]),True)
       
        val = i1.valGen.getValue(0.32, i1)
        self.assertEqual( sameFloatSequences(val, [3.20, 6.799, -8.199, -11.80]), True)
    
        val = i1.valGen.getValue(0.38, i1)
        self.assertEqual( sameFloatSequences(val, [3.799, 6.20, -8.80, -11.199]),True)
        # test configure()
        ends = [10., 20., 0., 0.]
        i1.valGen.generator.configure(lastVal = numpy.array(ends ))
        self.assertEqual( sameFloatSequences(i1.valGen.generator.valueRange, [10.0, 10.0, 5.0, 15.0]), True)
        self.assertEqual( sameFloatSequences(i1.valGen.getValue(0.5, i1), [5, 15, -2.5, -7.5]), True)
        self.assertEqual( sameFloatSequences(i1.valGen.getValue(0.32, i1),[3.2, 13.1999, -3.3999, -10.1999]), True)
        self.assertEqual( sameFloatSequences(i1.valGen.getValue(0.38, i1), [3.7999, 13.80, -3.10, -9.30]), True)
        self.assertEqual(i1.valGen.generator.nbvar , 4)


    def test_VarVectorInterpolator(self):

        val1 = [[1,1,1], ]
        val2 = [[6,6,6], [8,8,8], [10,10,10]] 

        kf1 = KF(0, val1)
        kf2 = KF(20, val2)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        bl = BehaviorList(None, None)
        ip = VarVectorInterpolator(None, None)
        bl.addBehavior(ip)
        i1.setValGenGenerator(bl)
        
        val= i1.valGen.getValue(0., i1)
        self.assertEqual( sameFloatSequences(val,val1), True)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual( sameFloatSequences(val, firstVal),True)
        
        val = i1.valGen.getValue(1., i1)
        self.assertEqual( sameFloatSequences(val,val2), True)
        lastVal = i1.valGen.generator.lastVal 
        self.assertEqual( sameFloatSequences(val, lastVal),True)
        
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(type(val), numpy.ndarray)
        self.assertEqual(sameFloatSequences(val.ravel() ,
                                      numpy.array([[ 3.5,  3.5,  3.5],
                                                     [ 4.5,  4.5,  4.5],
                                                     [ 5.5,  5.5,  5.5]], 'f').ravel()), True)
        val = i1.valGen.getValue(0.6, i1)
        self.assertEqual(type(val), numpy.ndarray)
        self.assertEqual(sameFloatSequences(val.ravel() ,
                                      numpy.array([[ 4.0,  4.0,  4.0 ],
                                                     [ 5.2,  5.2,  5.2,],
                                                     [ 6.4,  6.4,  6.4,]], 'f').ravel()), True)
        val = i1.valGen.getValue(0.67, i1)
        self.assertEqual(type(val), numpy.ndarray)
        self.assertEqual(sameFloatSequences(val.ravel() ,
                                      numpy.array([[ 4.35,  4.35,  4.35],
                                                     [ 5.69,  5.69,  5.69],
                                                     [ 7.03,  7.03,  7.03]], 'f').ravel()), True)
        
        # configure()
        i1.valGen.generator.configure(firstVal = val2, lastVal = val1)
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(type(val), numpy.ndarray)
        self.assertEqual(sameFloatSequences(val.ravel(),
                                            numpy.array([[3.5, 3.5, 3.5],
                                                           [4.5, 4.5, 4.5],
                                                           [5.5, 5.5, 5.5]], 'f').ravel()), True)

        val = i1.valGen.getValue(0.6, i1)
        self.assertEqual(type(val), numpy.ndarray)
        self.assertEqual(sameFloatSequences(val.ravel(),
                                            numpy.array([[3.0, 3.0, 3.0],
                                                           [3.8, 3.8, 3.8],
                                                           [4.6, 4.6, 4.6]], 'f').ravel()), True)

        val = i1.valGen.getValue(0.67, i1)
        self.assertEqual(type(val), numpy.ndarray)
        self.assertEqual(sameFloatSequences(val.ravel() ,
                                         numpy.array([[2.65, 2.65, 2.65],
                                                        [3.31, 3.31, 3.31],
                                                        [3.97, 3.97, 3.97]], 'f').ravel()), True) 
        
        
    def test_RotationInterpolator(self):

        val1 = [0.0, 0.0, 0.0, 1.0]
        val2 = [0.0, 1.0, 0.0, 0.0]
        kf1 = KF(0, val1)
        kf2 = KF(20, val2)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        ip = RotationInterpolator(None, None)
        i1.setValGenGenerator(ip)

        val= i1.valGen.getValue(0., i1)
        self.assertEqual( sameFloatSequences(val,val1), True)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual( sameFloatSequences(val, firstVal),True)
        
        val = i1.valGen.getValue(1., i1)
        self.assertEqual( sameFloatSequences(val,val2), True)
        lastVal = i1.valGen.generator.lastVal 
        self.assertEqual( sameFloatSequences(val, lastVal),True)


        val = i1.valGen.getValue(0.5, i1)
        newval = [0.0, 0.7071, 0.0, 0.7071]
        self.assertEqual( sameFloatSequences(val, newval),True)
        
        # test configure()
        i1.valGen.generator.configure(lastVal = newval)
        val = i1.valGen.getValue(1., i1)
        self.assertEqual( sameFloatSequences(val, newval),True)
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual( sameFloatSequences(val, newval),False)
        


    def test_FunctionInterpolator(self):
        func1 = FunctionInterpolator(function = simple)
        val = func1.getValue(0.5, None)
        self.assertEqual(val,0.5)

        func1 = FunctionInterpolator(function = adding)
        val = func1.getValue(0.5, None)
        self.assertEqual(val,1)

        #with posArgs,namedArgs
        func1 = FunctionInterpolator( function = (withargs,(2,),{'height':340,'width':20}))
        kf1 = KF(0,  5.0)
        kf2 = KF(20, 15.0)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        i1.setValGenGenerator(func1)

        val= i1.valGen.getValue(0., i1)
        self.assertEqual(val, 68000.0)
        firstVal = i1.valGen.generator.firstVal 
        self.assertEqual(5.0, firstVal)
        
        val = i1.valGen.getValue(1., i1)
        self.assertEqual(val, 204000.0)
        lastVal = i1.valGen.generator.lastVal 
        self.assertEqual(15.0, lastVal)
        

        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(val, 136000.0)
        
        i1.valGen.generator.configure(function = simple)
        val0 = i1.valGen.getValue(0, i1)
        self.assertEqual(val0,5.0)
        val1 = i1.valGen.getValue(1, i1)
        self.assertEqual(val1, 15.0)
        val = i1.valGen.getValue(0.5, i1)
        self.assertEqual(val, 10.0)

       

    def test_CompositeInterpolator(self):

        val1 = ([1,1,1], 1,)
        val2 = ([10,10,10], 10)
        kf1 = KF(0, val1)
        kf2 = KF(20, val2)
        valGen = ValueGenerator()
        i1 = Interval(kf1, kf2, valGen)
        ip = CompositeInterpolator(None, None,
                                   interpolators = (FloatVectorInterpolator, FloatScalarInterpolator))
        i1.setValGenGenerator(ip)

        val= ip.getValue(0.5, i1)
        #print "test Comp int;, val:", val
        self.assertEqual(len(val) , 2)
        
        self.assertEqual(sameFloatSequences(val[0], [5.5, 5.5, 5.5]), True)
        self.assertEqual(val[1], 5.5)
        val = ip.getValue(0.7, i1)
        self.assertEqual(len(val) , 2)
        self.assertEqual(sameFloatSequences(val[0], [7.299, 7.299, 7.299]), True)
        self.assertEqual(sameFloatSequences([val[1]], [7.299,]), True)

        #configure()
        ip.configure(interpolators = (FloatScalarInterpolator, IntScalarInterpolator,FloatVectorInterpolator),
                     firstVal = (1.0, 1,[1,1,1]), lastVal = (10.0, 10, [10,10,10]) )
        self.assertEqual(len(ip.interpolators) , 3)
        val= ip.getValue(0.5, i1)
        self.assertEqual(len(val) , 3)
        self.assertEqual(sameFloatSequences([val[0],] ,[5.5,]), True)
        self.assertEqual(val[1], 6)
        self.assertEqual(sameFloatSequences(val[2], [5.5, 5.5, 5.5]), True)
        
        val = ip.getValue(0.7, i1)
        self.assertEqual(len(val) , 3)
        self.assertEqual(sameFloatSequences([val[0],] ,[7.2999]), True)
        self.assertEqual(val[1], 7)
        self.assertEqual(sameFloatSequences(val[2], [7.299, 7.299, 7.299]), True)
                                   
if __name__ == '__main__':
    unittest.main()



from Scenario2.actor import Actor, CustomActor
from Scenario2.keyframes import ValueObject, ValueGenerator,KF, Interval, \
     AutoCurrentValueObject, KFAutoCurrentValue
from Scenario2.actions import Actions

class Object:

    def __init__(self):
        self.value = 0

    def get(self):
        return self.value

    def set(self, value):
        self.value = value


def getter(obj):
    return obj.get()


def test_001ActorConstructors():
    obj = Object()
    a = Actor('ObjectActor', obj)

    ca = CustomActor('CustomObjectActor', obj, setMethod='set',
                     getFunction=(getter, (), {}))
    
    

   
def test_008AutoCurrentValueObject():
    """
    test that this class return the current value of the associated object
    """
    class Object:
        def __init__(self, value):
            self.value = value
        def get(self):
            return self.value
        def set(self, value):
            self.value = value

    obj = Object(5)

    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))

    autoValue = AutoCurrentValueObject(actor)
    kf1 = KF(0, autoValue)
    assert isinstance(kf1.valueObject, AutoCurrentValueObject)
    assert kf1.getValue() == 5

    # change the object's value
    obj.set(8)

    # now the keyframe has to return the obejct's currnet value 8
    assert kf1.getValue() == 8


def test_008AutoCurrentValueObject():
    """
    test that this class return the current value of the associated object
    """
    class Object:
        def __init__(self, value):
            self.value = value
        def get(self):
            return self.value
        def set(self, value):
            self.value = value

    obj = Object(5)

    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))

    kf1 = KFAutoCurrentValue(0, actor)
    assert isinstance(kf1.valueObject, AutoCurrentValueObject)
    assert kf1.getValue() == 5

    # change the object's value
    obj.set(8)

    # now the keyframe has to return the obejct's currnet value 8
    assert kf1.getValue() == 8


def test_008intervalWithKFAutoCurrentValue():
    """
    Test KFAutoCurrentValue in interval.
    """
    class Object:
        def __init__(self, value):
            self.value = value
        def get(self):
            return self.value
        def set(self, value):
            self.value = value

    obj = Object(5.)
    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))
    kf1 = KFAutoCurrentValue(0, actor)
   
    kf2 = KF(10, 10.)
    valGen = ValueGenerator()
    i1 = Interval(kf1, kf2, valGen)

    from Scenario2.interpolators import FloatScalarInterpolator
    gen = FloatScalarInterpolator(None, None)
    i1.setValGenGenerator(gen)

    assert gen.firstVal == 5.
    assert gen.lastVal == 10.

    obj.set(8.)

    assert kf1.getValue() == 8.
    # after gettignthe value the interval using kf1 should
    # have updated first and last values
    assert gen.firstVal == 8.
   



##
## Author Michel Sanner Feb 2009
##
## Copyrights TSRI and M. Sanner
##

from Scenario2 import MaxKF
from Scenario2.director import MAADirector
from Scenario2.clipboard import Clipboard
from Scenario2.events import RemoveMAAEvent

class SequenceAnimator(MAADirector):
    """
    class storing a list of MultipleActorActions objects and allowing to play
    them in sequence and to rearrange their order.
    """

    def __init__(self):
        """
        SequenceAnimator object constructor.
        SequenceAnimatorObject <- SequenceAnimator()
        """
        MAADirector.__init__(self)
        self._animNB = None # will become a reference to the notebook widget
                            #of Pmv.customAnimation command


    def addMaa_cb(self, target, maa):
        """Callback to add a new maa to the SequeneAnimator. """
        self.addMAA(maa)


    def addMAA(self, maa):
        """
        Add maa at the end of SeqenceAnimator's maas list.
        val <- addMAA(maa)
        return value: returns True if maa is added successfully, False otherwise.
        """
        #end = self.endFrame
        #if end == 0 and len(self.maas)==0:
        #    position = 0
        #else:
        #    position = end + 1
        import types

        pos = self.endFrame
        maa.startOffset = 0
        maa.startFlag = "after previous"
        val = self.addMAAat( maa,  pos)
        return val


    def addMAAat(self, maa, pos, check=True):
        MAADirector.addMAAat(self, maa,pos, check)
        maa._animNB = self._animNB
        if hasattr(maa, "maas"):
            for _maa in maa.maas:
                _maa._animNB = self._animNB
            
 

    def moveMAA(self, maa, toPos):
        """ Move maa  to specified position (toPos). Update the positions of maas
        in the director's maas list
        None <- moveMaa( maa, toPos)
        """
        # maa is a (maa, position) tuple
        fromPos = self.maas.index(maa)
        
        # remove it
        self.maas.remove(maa)
        self.maas.insert(toPos, maa)

        # renumber positions
        pos = 0
##        pos = 0
##         for i in range(len(self.maas)):
##             self.maas[i][1] = pos
##             pos += self.maas[i][0].lastPosition + 1
        self.updatePositions(0, pos)


    def removeMAA(self, maa, position):
        """
        remove maa from sequence animator and renumber the starting position of
        all MAA after the removed one.

        None <- removeMAA(maa, position)
        arguments:
           maa: instance of MAA;
           position: the specified MAA starting position (offset) in the MAADirector.
        Event:
           generates a RemoveMAA event
        """

        # remove maa
        index = self.maas.index([maa, position])
        self.maas.pop( index )

        # MS: This is done by self.updatePositions
        # update director's endFrame
        #self.endFrame -= maa.lastPosition
        
        # renumber positions
        pos = position
        if index < len(self.maas):
            nextmaa = self.maas[index]
            startFlag =  nextmaa[0].startFlag
            if startFlag == "with previous" :
                if index > 0:
                    pos = self.maas[index-1][1]
                else:
                    pos = 0
            else: # "after previous"
                if index > 0:
                    if maa.startFlag != "after previous":
                        pos = self.maas[index-1][1] + self.maas[index-1][0].lastPosition + 1
                else:
                    pos = 0

##         for i in range(index, len(self.maas)):
##             self.maas[i][1] = pos
##             pos += self.maas[i][0].lastPosition + 1
        
        self.updatePositions(index, pos)
            
        # create and dispatch event
        self.dispatchEvent(RemoveMAAEvent(maa))


    def updatePositions(self, index, position):
        """Recompute positions of maas.
        None <- updatePositions(index, position)

        arguments:
           index - the maa index (in self.maas list) from which positions
                   should be recomputed
           position - frame number at which self.maas[index] will start

        """

        nbmaas = len(self.maas)

        #print "updatePositions", "index:", index, "position" , position

        if index >= nbmaas:
            self.updateEndFrame()
            return
        # find max end position of the maas upto the specified maa index
        maxpos = 0
        for i in range(index+1):
            maa, pos = self.maas[i]
            end = pos + maa.lastPosition
            if end > maxpos:
                maxpos = end
                
        # move back to first real start
        if self.maas[index][0].startFlag== 'with previous':
            while index > 0 and self.maas[index][0].startFlag== 'with previous':
                index -= 1
       
        

        # remember which MAA was the last one seen that was nor ith previous
        lastStart = index

        # set start position for maa
        maa, startFrame = self.maas[index]
        self.maas[index][1] = position + maa.startOffset # set starting position
        # pos will be the starting position of the next MAA
        pos = position + maa.startOffset + maa.lastPosition + 1
        maxpos = max(maxpos, pos)
        # loop over remaining MAAs in list 
        for i in range(index+1,  nbmaas):
            maa, startFrame = self.maas[i] 
            startFlag = maa.startFlag
            if startFlag == "with previous":
                # set start to stat of lastStart
                self.maas[i][1] = self.maas[lastStart][1] + maa.startOffset
                
                #set pos to the last frame of current maa + 1
                pos = self.maas[lastStart][1] + maa.startOffset + \
                           maa.lastPosition + 1
                # set maxpos to the maxinum nbumber of frames covered by MAAs
                # under this parent
                maxpos = max( maxpos, pos)
                
            elif startFlag == "after previous":
                # set the starting position to pos
                self.maas[i][1] = pos + maa.startOffset
                # increment pos to the next available frame
                pos += maa.startOffset + maa.lastPosition + 1
                maxpos = max( maxpos, pos)
                # remember that this MAA was not with previous
                lastStart = i
        # set the sequenceAnimator endFrame to the the first unused frame
        if nbmaas > 0:
            #self.endFrame = pos #self.maas[-1][1]+ self.maas[-1][0].lastPosition
            self.endFrame = maxpos
        else:
            self.endFrame = 0




try:
    from Scenario2.datatypes import FloatType
    from Scenario2.interpolators import FloatScalarInterpolator
    from Scenario2.director import Director
    from Scenario2.actor import Actor, CustomActor
    from Scenario2.keyframes import ValueObject, ValueGenerator, KF, Interval
except:
    pass

class foo:

    def __init__(self):
        self.a= 2.0

    def set(self, value):
        self.a = value

    def get(self):
        return self.a
    

def test_001ActorInDirector():

    d = Director()
   
    f = foo()

    # create an actor
    actor = Actor('test', f, 0.5, FloatType, FloatScalarInterpolator)
    assert len(actor.actions.keyframes) == 1
    assert actor.actions.keyframes[0].getValue() == 0.5

    # create a director
    actor = d.addActor(actor)
    assert actor._maa() == d

    # run the animation
    print "Run keyframe 0"
    d.run()


def test_002CustomActorInDirector():

    d = Director()

    f = foo()

    # create an actor
    actor = CustomActor('test', f, 0.5, FloatType, FloatScalarInterpolator,
                        setMethod='set', getFunction=(f.get, (), {}))
    assert len(actor.actions.keyframes) == 1
    assert actor.actions.keyframes[0].getValue() == 0.5

    # add actor to director
    actor = d.addActor(actor)
    
    # run the animation
    print "Run keyframe 0"
    d.run()


    # add Interval
    kf1 = KF( 10, 25 )
    kf2 = KF( 20, 35 )
    i1 = Interval(kf1, kf2, )
    actor.addIntervals( [i1] )
    assert d.endFrame == 20
    d.run()


## actor.setKeyframe( 10, 10.5)
## print "Run keyframe 0 and 1"
## d.run()

## # add a keyframe at 15
## actor.setKeyframe( 20, 10.5)
## actor.setKeyframe( 25, 5.5)
## assert len(actor.valueGenerators)==4
## assert len(actor.keyframes.kfs)==4
## print "Run keyframe 0,1,2,3"
## d.run()

## # stop interpolation between keyframe 1 and 2 (i.e interval 1)
## actor.valueGenerators[1].configure(active = False)
## #actor.setValueGenerator( None, segment=1)
## print "Run keyframe 0,1 and 2,3"
## d.run()




##
## Author Michel Sanner Feb 2009
##
## Copyrights TSRI and M. Sanner
##

import Tkinter, Pmw

from Scenario2.clipboard import AddMAAToClipBoardEvent, RemoveMAAFromClipBoardEvent
from Scenario2.clipboard import Clipboard

from mglutil.util.callback import CallBackFunction

class ClipboardGUI:
    """
    class displaying the list of MultipleActorActions stored in
    a clipboard object.

    This object handles AddMAAToClipBoardEvent and RemoveMAAFromClipBoardEvent
    events dispatched by the standard scenario event handler
    """

    def __init__(self, clipboard, master=None):
        """
        ClipboardGUI constructor
        
        ClipboardGUIObject <- ClipboardGUI(clipboard, master=None)

        clipboard is an instance of a Clipboard object
        master is a Tk container object. If master is None a Tkinter.Toplevel
        object is created

        This object registers interest in AddMAAToClipBoardEvent and
        RemoveMAAFromClipBoardEvent events with the Clipboard object
        """
        assert isinstance(clipboard, Clipboard)

        self.clipboard = clipboard

        self.master = master

        clipboard.registerListener(AddMAAToClipBoardEvent,
                                   self.addMaaEvent_cb)
        clipboard.registerListener(RemoveMAAFromClipBoardEvent,
                                   self.removeMaaEvent_cb)
        self.createGUI()
        self.refreshGUI()


    def addMaaEvent_cb(self, event=None):
        self._addMAA(event.object)


    def removeMaaEvent_cb(self, event=None):
        self.clearGUI()
        self.refreshGUI()


    def createGUI(self):
        """
        Create a ScrolledFrame to old MAA entries
        """
        if self.master is None:
            self.master = master = Tkinter.Toplevel()
            self.ownsMaster = True
        else:
            self.ownsMaster = False

        # create container
        w = self.MAAContainer = Pmw.ScrolledFrame(
            self.master, labelpos = 'nw', label_text = 'Animations Clipboard',
            usehullsize = 0, hull_width = 40, hull_height = 200)

        w.pack(padx = 5, pady = 3, fill = 'both', expand = 1)
       
        self.maaw = {}   # maa: maa entry widget


    def showMenu_cb(self, maa, event=None):
        """
        configure callback of menu entried with the current orientation and post
        menu
        """
        menu = Tkinter.Menu(self.master, title = "Animation")
        from Scenario2 import addTargetsToMenu
        addTargetsToMenu(menu, maa)
        menu.add_command(label = "Delete",
                  command=CallBackFunction(self.clipboard.removeMaa, maa) )
        menu.add_command(label="Dismiss")
        menu.post(event.x_root, event.y_root)


    def clearGUI(self):
        for maa, entry in self.maaw.items():
            entry.destroy()


    def _addMAA(self, maa):
         """
         Create an entry for a MAA
         """

         master = self.MAAContainer.interior()
       
         self.maaw[maa] = frame = Tkinter.Frame(master)
         
         b = Tkinter.Button(master=frame ,compound='left', text=maa.name,
                            command=maa.run, width=25, height=2)

         b.name = maa.name
         # create the button menu
         b.bind('<Button-3>', CallBackFunction( self.showMenu_cb, maa))
       
         b.pack(side='top')
         frame.pack()


    def refreshGUI(self):

        for maa in self.clipboard.maas:
            self._addMAA(maa)




##
##  Author Michel F. Sanner Jan 2009
##
class ValueObject:
    """Object holding a value associated with a keyframe
    """

    def __init__(self, value=None, cmp=None):
        """Constructor of ValueObject
        None <- ValueObject(value=None, cmp=None)
        arguments: value  -  a value associated with a keyframe
                   cmp - a function used to compare the value of this ValueObject with
                         the value of another ValueObject.
        """
        self.value = value
        self.cmp = cmp
        
    def getValue(self):
        """
        value <- getValue()
        returns self.value
        """
        return self.value

    def setValue(self, value):
        """
        None <- setValue(value)
        Set the value atttribute
        """
        self.value = value

    def equal(self, valueObject):
        """ 
        bool <- equal(valueObject)
        compare the value stored in self.value with
        the value of specified ValueObject.
        Returns True or False
        """
        if self.cmp:
            return self.cmp(self.value, valueObject.value)
        else:
            return self.value == valueObject.value


class AutoCurrentValueObject(ValueObject):
    """
    ValueObject class that will get and set it's value autoamtically to the
    current value of the actor's animated object when it plays.
    The actor has to be able to retrieve the
    current value from the object it is driving.
    """
    def __init__(self, actor):
        """
        None <- AutoCurrentValueObject(actor)
        argument:
          actor - an Actor instance, has to have getValueFromObject method 
        """
        ValueObject.__init__(self)
        assert actor.getFuncTuple
        self.actor = actor

    def getValue(self):
        """
        value <- getValue()
        Returns the current value of the actor's object associated with this ValueObject
        """
        return self.actor.getValueFromObject()


class ValueFromFunctionObject(ValueObject):
    """
    ValueObject object that will get a value from a specified function
    when it plays.
    """

    def __init__(self, function, *args, **kw):
        ValueObject.__init__(self)
        assert callable(function)
        self.function = function
        self.args = args
        self.kw = kw
        
    def getValue(self):
        return self.function(*self.args, **self.kw)
    

class ValueGenerator:
    """
    Object producing interpolating values over an interval.
    The generator has to be callable and has to return a value for a given
    input value. This input value is usually a fraction of an interval ranging
    from 0. at the begining of the interval to 1. at the end of the interval. 
    """
    def __init__(self, generator=None):
        self.generator = None
        if generator is not None:
            self.setGenerator(generator)

    def setGenerator(self, generator):
        assert callable(generator)
        self.generator = generator
        
    def getValue(self, percentage, interval, **kw):
        if self.generator is None:
            raise ValueError("ValueGenerator has no generator set yet")
        else:
            return self.generator( *(percentage, interval), **kw)

    def clone(self):
        """
        This method returns a copy of the ValueGenerator object.
        If the generator has a .clone() method this method will be used to clone
        this ValueGenerator. Else the cloned ValueGenerator will be crated by
        pointing to the same callable as the ValueGenerator beign cloned.
        """
        try:
            generator = self.generator.clone()
        except AttributeError:
            generator = self.generator

        nvg = self.__class__(generator)
        return nvg


class LinkedList:
    """This mixin class adds 2 attributes nxtLinked and prvLinked and proviceds support for adding a linked copy using linkedCopy() and retrieving all linked copies 
"""

    def __init__(self):
        self.nxtLinked = None # next linked keyframe
        self.prvLinked = None # previous linked keyframe


    def linkedCopy(self, newLinkedObject):
        """update links to add the linkedObject
"""
        #find last linked keyframe
        lastkf = self
        while lastkf.nxtLinked is not None:
            lastkf = lastkf.nxtLinked

        # build links between last and linkedObject
        newLinkedObject.prvLinked = lastkf
        lastkf.nxtLinked = newLinkedObject


    def getAllLinked(self):
        """
        return all linked keyframes, self is always first in the returned list
        """

        kf = self
        linked = [kf,]
        while kf.nxtLinked is not None:
            kf = kf.nxtLinked
            linked.append( kf )

        kf = self.prvLinked
        if kf is None:
            return linked

        linked.append( kf )
        while kf.prvLinked is not None:
            kf = kf.prvLinked
            linked.append( kf )
        return linked
        

    def isLinked(self):
        return self.nxtLinked is not None or self.prvLinked is not None
   

class KF(LinkedList):
    """Keyframe object. This object associates a frame (i.e. time) with a value
stored in a ValueObject object.
"""
    def __init__(self, position, value):

        LinkedList.__init__(self)

        self.pos = position
        if isinstance(value, ValueObject):
            self.valueObject = value
        else:
            self.valueObject = ValueObject(value)

        self.leftInterval = None # interval object before KF
        self.rightInterval = None # interval object after KF


    def getValue(self, frame=None, interval=None):
        return self.valueObject.getValue()

        
    def setValue(self, value):
        self.valueObject.setValue(value)

        # for this keyframe and all linked ones update valGens
        for k in self.getAllLinked():
            # update value generators for left interval
            inter = k.leftInterval
            if inter:
                if inter.valGen.generator:
                    inter.valGen.generator.configure(lastVal=value)

            # update value generators for right interval
            inter = k.rightInterval
            if inter:
                if inter.valGen.generator:
                    inter.valGen.generator.configure(firstVal=value)
                    
        
    def linkedCopy(self, position):
        """update links to add the linkedObject
"""
        newkf = KF(position, self.valueObject)
        LinkedList.linkedCopy(self, newkf)
        return newkf


    def copy(self, position):
        """
        create a deep copy of a keyframe
        """
        return self.__class__(position, self.getValue())


    def equal(self, kf):
        """
        return True if kf and self have same position and same value
        """
        return self.pos==kf.pos and \
               ( (self.valueObject==kf.valueObject) or (self.valueObject.equal(kf.valueObject)) )
    

class KFAutoCurrentValue(KF):
    """Keyframe object. This object associates a frame (i.e. time) with a value
    returned or generated by AutoCurrentValueObject object.
"""
    def __init__(self, position, value):
        """
        Constructor of KFAutoCurrentValue obejct

        obj <- KFAutoCurrentValue(position, value)

        position is an integer describing the time for this keyframe
        value can be a AutoCurrentValueObject instance or an object for which
        a AutoCurrentValueObject can be created
        """
        LinkedList.__init__(self)

        self.pos = position
        if isinstance(value, AutoCurrentValueObject):
            self.valueObject = value
        else:
            self.valueObject = AutoCurrentValueObject(value)

        self.leftInterval = None # interval object before KF
        self.rightInterval = None # interval object after KF


    def getValue(self, value=None):
        if value is None:
            value = self.valueObject.getValue()
        #print "kf: ", self, self.pos, value
        inter = self.rightInterval
        if inter:
            gen = inter.valGen.generator
            if gen:
                if hasattr(gen, 'configure'):
                    try:
                        gen.configure(firstVal=value, active = True)
                        inter.active = True
                    except:
                        gen.configure(active = False)
                        inter.active = False
        inter = self.leftInterval
        if inter:
            gen = inter.valGen.generator
            if gen:
                if hasattr(gen, 'configure'):
                    try:
                        gen.configure(lastVal=value, active = True)
                        inter.active = True
                    except:
                        gen.configure(active = False)
                        inter.active = False
        return value

    def copy(self, position):
        """
        create a deep copy of a keyframe
        """
        return KFAutoCurrentValue(position, self.valueObject.actor)


class KFValueFromFunction(KF):
    """Keyframe object. This object associates a frame (i.e. time) with a value
    returned or generated by ValueFromFunctionObject object.
"""
    def __init__(self, position, function):
        """
        Constructor of KFAutoCurrentValue obejct

        obj <- KFValueFromFunction(position, function)

        position - an integer describing the time for this keyframe
        function - either a ValueFromFunctionObject instance or a callable function  for which
        a ValueFromFunctionObject can be created
        """
        KF.__init__(self, position, None)

        if isinstance(function, ValueFromFunctionObject):
            self.valueObject = function
        else:
            self.valueObject = ValueFromFunctionObject(function)


    def getValue(self):
        value = self.valueObject.getValue()
        inter = self.rightInterval
        if inter:
            gen = inter.valGen.generator
            if gen:
                if hasattr(gen, 'configure'):
                    gen.configure(firstVal=value)
        inter = self.leftInterval
        if inter:
            gen = inter.valGen.generator
            if gen:
                if hasattr(gen, 'configure'):
                    gen.configure(lastVal=value)

        return value



    def copy(self, position):
        """
        create a deep copy of a keyframe
        """
        return self.__class__(position, self.valueObject.function)

            
class Interval(LinkedList):
    """Class representing an interval to the right of the keyframe k1
"""
    def __init__(self, k1=None, k2=None, valGen=None, generator=None, **kw):
        """
        Interval constructor

        interval <- Interval(k1=None, k2=None, valGen=None, generator=None)

        If k1 and k2 are specified that has to be instances of KF objects
        If valGen is specified it has to be an instance of ValueGenerator
        if valGen is NOT specified and generator is specified, a ValueGenerator
        object is built and generator is set as the generator for this
        ValueGenerator
        """
        LinkedList.__init__(self)

        self.kf1 = None
        self.kf2 = None
        self.getValue = self._getValue0KF

        if valGen is None:
            self.valGen = ValueGenerator(generator)
        else:
            assert isinstance(valGen, ValueGenerator)
            self.valGen = valGen


        #self.setKF1(k1)
        #self.setKF2(k2)
        self.setKeyFrames(k1, k2)

        self.active = True # when false this interval deos not set values

        self.data = kw.copy() # used to propagate data such as rotation angle


    def setActive(self, value):
        assert value in [True, False, 0, 1]
        self.active = bool(value)


    def setGetValueMethod(self):
        """
        select getValue method depending on the keyframes that are defined
        """
        if self.kf2 is None and self.kf1 is None:
            self.getValue = self._getValue0KF

        elif self.kf1 is not None and self.kf2 is None:
            self.getValue = self._getValue1KF
            
        elif self.kf1 is not None and self.kf2 is not None:
            self.getValue = self._getValue2KF
        else:
            self.getValue = self._getValue0KF


    def setKF1(self, k1):
        """
        set first keyframe
        """
        assert k1 is None or isinstance(k1, KF)
        self.kf1 = k1
        if k1 is not None:
            k1.rightInterval = self

            val  = k1.getValue()
            if val is not None:
                if self.valGen.generator:
                    gen = self.valGen.generator
                    if hasattr(gen, 'configure'):
                        gen.configure(firstVal=val)

        self.setGetValueMethod()


    def setKF2(self, k2):
        """
        set second keyframe
        """
        assert k2 is None or isinstance(k2, KF)
        self.kf2 = k2
        if k2 is not None:
            k2.leftInterval = self

            val  = k2.getValue()
            if val is not None:
                if self.valGen.generator:
                    gen = self.valGen.generator
                    if hasattr(gen, 'configure'):
                        gen.configure(lastVal=val)

        self.setGetValueMethod()


    def setKeyFrames(self, k1, k2):
        """set both keyframes """
        assert k1 is None or isinstance(k1, KF)
        assert k2 is None or isinstance(k2, KF)
        self.kf1 = k1
        self.kf2 = k2
        if k1 is not None and k2 is not None:
            k1.rightInterval = self
            k2.leftInterval = self
            if self.valGen.generator:
                gen = self.valGen.generator
                if hasattr(gen, 'configure'):
                    val1  = k1.getValue()
                    val2 =  k2.getValue()
                    if val1 is not None and val2 is not None:
                        gen.configure(firstVal=val1, lastVal=val2)
                    elif val1 is not None:
                        gen.configure(firstVal=val1)
                    elif val2 is not None:
                        gen.configure(lastVal=val2)
            self.setGetValueMethod()
        elif k1 is not None:
            self.setKF1(k1)
        elif k2 is not None:
            self.setKF2(k2)
        

    def setValGen(self, valGen):
        """
        set the interval's value generator
        """

        assert valGen is None or isinstance(valGen, ValueGenerator)
        self.valGen = valGen


    def setValGenGenerator(self, generator):
        """
        set the value generator function in the interval's value generator object
        """

        self.valGen.setGenerator(generator)
        
        # set the keyframes to for the generator to be configured
        if self.kf1:
            generator.configure(firstVal=self.kf1.getValue())
        if self.kf2:
            generator.configure(lastVal=self.kf2.getValue())


    def linkedCopy(self, position):
        """Returns a linked interval with linked keyframes and pointing to the same calue generator than the original interval.
"""
        oldpos1 = self.kf1.pos
        k1l = self.kf1.linkedCopy(position)
        if self.kf2 is not None:
            k2l = self.kf2.linkedCopy(self.kf2.pos + (position-oldpos1))
        else:
            k2l = None
        newInt = Interval( k1l, k2l, self.valGen )

        LinkedList.linkedCopy(self, newInt)

        return newInt


    def _getValue2KF(self, frame):
        """intervals with 2 keyframes return a value generated by the interval value generator"""
        if self.active:
            first = self.kf1.pos
            nbFrames = float(self.kf2.pos - first) + 1
            #print 'AAA', frame, nbFrames
            if frame==self.kf2.pos:
                fraction = 1.0
            else:
                fraction = (frame-first)/(nbFrames-1)

            #try:
            #    return self.valGen.getValue(fraction, h, dy)
            #except:
            return self.valGen.getValue( fraction, self )
        else:
            return None

    
    def _getValue1KF(self, frame):
        """open ended intervals return value of starting keyfram"""
        if self.active:
            return self.kf1.getValue()
        else:
            return None

            
    def _getValue0KF(self, frame):
        """intervals with no keyframes return None"""
        return None

    
    def getValue(self, frame):
        """
        Retuns the value for the given frame.
        The value is only returned if the interval is active.
        This frame has to be withing the 2 keyframes of the interval,
        else we raise a ValueError.
        If the interval has no second keyframe, we return the value of the
        first keyframe.
        """
        # this method is overriden when keyframes are set
        pass
    



from Scenario2.keyframes import ValueObject, ValueGenerator,KF, Interval, AutoCurrentValueObject , KFAutoCurrentValue
from Scenario2.interpolators import FloatScalarInterpolator

def test_001KFvalue1():
    """
    test ValueObject object contructor, setting and getting its value
    """
    # create object with no value
    obj = ValueObject()
    assert obj.getValue()==None

    obj.setValue(6)
    assert obj.getValue()==6

    obj.setValue('Hello World')
    assert obj.getValue()=='Hello World'

    # create object with passing value
    obj = ValueObject(6)
    assert obj.getValue()==6

    obj= ValueObject('Hello World')
    assert obj.getValue()=='Hello World'


def test_002ValueGenerator1():
    """
    test the creation of a value generator object
    """
    # create object with no value
    obj = ValueGenerator()
    try:
        obj.getValue(0.5, None)
    except ValueError:
        pass

    def f(percent, interval=None):
        return percent*2

    obj.setGenerator(f)
    assert obj.getValue(2, None)==4


def test_002ValueGenerator2():
    # create object with no value
    obj = ValueGenerator()

    # clone the ValueGenerator with no generator
    obj1 = obj.clone()
    assert isinstance(obj1, obj.__class__)
    assert obj.generator == obj1.generator

    def f(percent):
        return percent*2

    obj.setGenerator(f)
    assert obj.generator == f

    #clone the ValueGenerator with function
    obj1 = obj.clone()
    assert isinstance(obj1, obj.__class__)
    assert obj.generator == obj1.generator

    #clone the ValueGenerator with generator support clone() method
    gen = FloatScalarInterpolator(None, None)
    obj.setGenerator(gen)
    assert obj.generator == gen
    obj1 = obj.clone()
    assert isinstance(obj1, obj.__class__)
    assert obj.generator != obj1.generator
    assert obj.generator == gen
    

def test_003KF1():
    """
    test KF contructor
    """

    # KF requires position when created
    try:
        kf = KF()
    except TypeError:
        pass

    # with position but no value
    kf = KF(0, None)
    assert kf.pos == 0
    # we expect AttributeError if we try to get the value
    try:
        assert kf.getValue() == None
    except AttributeError:
        pass
    
    # with position and value is a simple value
    kf = KF(5, 7)
    assert kf.pos == 5
    assert kf.getValue() == 7

    # with position and value is a ValueObject
    kv = ValueObject('Hello')
    kf = KF(5, kv)
    
    assert kf.pos == 5
    assert kf.getValue() == 'Hello'

    # create second keyframe with same ValueObject
    kf1 = KF(12, kv)
    assert kf1.pos == 12
    assert kf1.getValue() == 'Hello'

    # make sure that if we change the value in kv both KF see the change
    kv.setValue('Bye')
    assert kf.getValue()=='Bye'
    assert kf1.getValue()=='Bye'


def test_004linkedKF1():
    """
    test that linked KFs are build properly and can be retrieved
    """
    # get a linked copy of a keyframe
    kf1 = KF(10, 10)
    kf2 = kf1.linkedCopy(15)
    assert kf2.pos == 15
    assert kf2.getValue() == 10
    assert kf1.nxtLinked == kf2
    assert kf1.prvLinked == None
    assert kf2.prvLinked == kf1
    assert kf2.nxtLinked == None

    # make sure either of the 2 KF can set the value
    kf1.setValue( 32 )
    assert kf1.getValue() == 32
    assert kf2.getValue() == 32

    kf2.setValue( 28 )
    assert kf1.getValue() == 28
    assert kf2.getValue() == 28

    # get a linked copy of kf2
    kf3 = kf2.linkedCopy(20)

    # get another copy of kf1
    kf4 = kf1.linkedCopy(16)

    # get all linked KFs
    lkf = kf1.getAllLinked()
    assert len(lkf)==4
    assert kf1==lkf[0]
    assert kf2 in lkf
    assert kf3 in lkf
    assert kf4 in lkf

    lkf = kf2.getAllLinked()
    assert len(lkf)==4
    assert kf2==lkf[0]
    assert kf1 in lkf
    assert kf3 in lkf
    assert kf4 in lkf

    lkf = kf3.getAllLinked()
    assert len(lkf)==4
    assert kf3==lkf[0]
    assert kf1 in lkf
    assert kf2 in lkf
    assert kf4 in lkf

    lkf = kf4.getAllLinked()
    assert len(lkf)==4
    assert kf4==lkf[0]
    assert kf1 in lkf
    assert kf2 in lkf
    assert kf3 in lkf


def test_005KFcopy():
    """
    test creating a copy of a keyframe
    """
    kf1 = KF(10, 10)
    kf2 = kf1.copy(15)
    assert kf2.pos == 15
    assert kf2.getValue() == 10
    assert kf2.valueObject != kf1.valueObject
    kf2.setValue(30)
    assert kf2.getValue() == 30
    assert kf1.getValue() == 10
   
    
def test_005Interval1():
    """
    test Interval constructor
    """
    i1 = Interval()
    assert i1.kf1 is None
    assert i1.kf2 is None
    assert isinstance(i1.valGen, ValueGenerator)

    kf1 = KF(0, 0)
    i1 = Interval(kf1)
    assert i1.kf1 == kf1
    assert i1.kf2 is None
    assert isinstance(i1.valGen, ValueGenerator)

    kf2 = KF(10, 10)
    i1 = Interval(kf1, kf2)
    assert i1.kf1 == kf1
    assert i1.kf2 == kf2
    assert isinstance(i1.valGen, ValueGenerator)

    vg = ValueGenerator()
    i1 = Interval(kf1, kf2, vg)
    assert i1.kf1 == kf1
    assert i1.kf2 == kf2
    assert i1.valGen == vg
    

def test_006intervalLinked1():
    """
    test creating linked Interval 
    """
    kf1 = KF(0, 10)
    kf2 = KF(10, 0)
    i1 = Interval(kf1, kf2)
    i2 = i1.linkedCopy(20)
    
    assert i2.kf1.pos == 20
    assert i2.kf2.pos == 30
    assert i2.valGen == i2.valGen

    i3 = i2.linkedCopy(40)
    assert i3.kf1.pos == 40
    assert i3.kf2.pos == 50
    assert i3.valGen == i2.valGen
    assert i3.valGen == i1.valGen

    li = i1.getAllLinked()
    assert len(li) == 3
    assert li[0] == i1
    assert i2 in li
    assert i3 in li

    li = i2.getAllLinked()
    assert len(li) == 3
    assert li[0] == i2
    assert i1 in li
    assert i3 in li
    

def test_007intervalValGen():
    kf1 = KF(0, 0)
    kf2 = KF(10, 10)
    valGen = ValueGenerator()
    i1 = Interval(kf1, kf2, valGen)

    gen = FloatScalarInterpolator(None, None)
    i1.setValGenGenerator(gen)

    assert i1.getValue(0) == 0.0
    assert i1.getValue(10) == 10.0
    assert i1.getValue(5) == 5.0

    i1 = Interval(kf1, None, valGen)
    assert i1.getValue(10) == 0.0
    assert i1.getValue(100) == 0.0



def test_008KFAutoCurrentValue():
    """Test creation KFAutoCurrentValue keyframe with AutoCurrentValueObject object """
    class Object:

        def __init__(self, val):
            self.value = val

        def get(self):
            return self.value

        def set(self, value):
            self.value = value

    obj = Object(5)
    from Scenario2.actor import CustomActor
    actor = CustomActor('obj', obj, setMethod='set',
                        getFunction=(obj.get, (), {}))

    autoval = AutoCurrentValueObject(actor)
    kf = KFAutoCurrentValue(0, autoval)
    assert kf.getValue() == 5
    



from mglutil.gui.BasicWidgets.Tk.player import Player
from Scenario2.events import PlayerStartEvent, PlayerStopEvent
from Scenario2.director import MAADirector
from mglutil.util.callback import CallbackFunction
import Tkinter, Pmw
import sys, os
import tkFileDialog
from time import strftime
import shutil, tarfile

class PlayerControls(Player):
    """Subclasses Player widget. Plays and records animations of the
    specified director object.
    PlayerControlsObject <- PlayerControls (director, kw)
    """
    def __init__(self, director, **kw):
##                         master=None, root=None,
##                         height=80,width=200,
##                         currentFrameIndex=0,
##                         startFrame=0, 
##                         endFrame=0,
##                         maxFrame=0,
##                         stepSize=1, 
##                         playMode=0,
##                         ##afterDelay=50,
##                         titleStr='Player',
##                         gotoStartfile = 'go_to_start.gif',
##                         gotoEndfile = 'go_to_end.gif',
##                         ff_revfile = 'ff_rev.gif',
##                         ff_fwdfile = 'ff_fwd.gif',
##                         stopfile = 'stop.gif',
##                         playfile = 'play_fwd.gif',
##                         playRevfile = 'play_rev.gif',
##                         chmodfile = 'chmod.gif',
##                         closefile = 'close.gif',
##                         iconpath = None,
##                         counter = 1,
##                         form2=1, gui=1,framerate=15., hasSlider=False):

        assert isinstance(director, MAADirector)
        self.director = director
        kw['gui'] = 1
        kw['form2'] = 1
        kw['startFrame'] = 0
        kw['maxFrame'] = kw['endFrame'] = director.endFrame
        kw['buttonMask'] = {'setanimB':False, 'closeB':False}
        kw['currentFrameIndex'] = -1
        Player.__init__(self, **kw)
        if self.hasSlider:
            slider = self.form.descr.entryByName['slider']['widget']
            sliderkw = {'from': self.startFrame-1, 'to': self.maxFrame+1}
            slider.configure(sliderkw)
            slider.set(self.currentFrameIndex)
        self.form.ent2.delete(0,'end')
        self.form.ent2.insert(0, 'start')
        self.recording = False
        if sys.platform == 'darwin':
            self.codecType = 'mpeg1video'
        else:
            self.codecType = 'mpeg2video'
        self.filename = 'movie.mpg'
        self.imagefilename = ""
        self.cameraSize=Tkinter.StringVar()
        self.cameraSize.set("current")
        root = self.root
        screenw, screenh = root.winfo_screenwidth(), root.winfo_screenheight()
        res = [(320, 240), (640, 480), (1280, 720), (1920, 1080)]
        self.resolutions = {}
        for w,h in res:
            if w <= screenw and h <=screenh:
                self.resolutions["%dx%d"%(w,h)] = (w, h)
        nb = self.director._animNB
        if nb:
            self.vf=nb().pmv
        else:
            self.vf=None
        self.recordImages = False
        self.recordDialog = self.createRecordDialog()
        


    def createRecordDialog(self):
        self.balloon = Pmw.Balloon(self.master)
        dialog = Pmw.Dialog(
            self.master, buttons=['Record', 'Cancel'], defaultbutton='Record',
            title="Record Options")
        dialog.withdraw()
        # create a frame to hold group to force setting orient and rendering
        frame = Tkinter.Frame(dialog.interior())

        bn = Tkinter.Button(frame, text="Save As:", command = self.browseFile_cb,
                            width= 0, height=0)
        self.balloon.bind(bn, "Opens file browser")
        bn.grid(row=0, column=0,sticky='ew')
        
        filename = 'movie_'+strftime("%m-%d-%Y_%H:%M")+'.mpg'
        self.filenamew = fn = Pmw.EntryField(frame, command = self.getFileName_cb,
                                             entry_width = 35,
                                             value = filename)
        fn.grid( row=0, column = 1,sticky ='w') #columnspan=2)
        self.balloon.bind(fn, "type filename" )
        self.codecw = Pmw.RadioSelect(frame, labelpos='w', label_text="Codec:",
                                      selectmode='single', orient='horizontal',
                                      buttontype='radiobutton', command=self.setCodec_cb)
        for text in ['mpeg1video', 'mpeg2video']:
            self.codecw.add(text)

        self.codecw.grid( row=1, column=0, columnspan=2, sticky='ew')
        self.cameraSizeMenub = None
        row = 1
        if self.vf:
            listitems = self.resolutions.keys()
##             csize = Pmw.ComboBox(frame, labelpos = 'w',
##                                 label_text = 'Camera Size:',
##                                 #entryfield_validate = self.entryValidate,
##                                 entryfield_value="current",
##                                 scrolledlist_items=listitems, 
##                                 #fliparrow=1,
##                                  history=False,
##                                  selectioncommand=self.setCameraSize_cb)
            
##             csize.grid( row=2, column=0, columnspan=2, sticky='ew')
            self.cameraSizeMenub = mb= Tkinter.Menubutton(frame, text='Select camera size...',
                           underline=0, relief='raised')
            #mb('<Button-1>', toggleMenu, '+')
            row = row+1
            mb.grid( row=row, column=0, columnspan=2, sticky='ew')
            mb.menu = Tkinter.Menu(mb)
            for s in self.resolutions.keys():
                mb.menu.add_radiobutton(label=s,
                                        var=self.cameraSize,
                                        value = s,
                                        command = self.setCameraSize_cb)
            mb.menu.add_radiobutton(label="current", var=self.cameraSize,
                                    value="current",
                                    command = self.setCameraSize_cb)
            mb['menu'] = mb.menu
            self.balloon.bind(mb, "Left-Click to select camera size")

        bn = Tkinter.Button(frame, text="Save frames\nas image files:",
                            command = CallbackFunction(self.browseFile_cb, "IMG"),
                            width= 0, height=0)
        self.balloon.bind(bn, "Opens file browser")
        row = row+1
        bn.grid(row=row, column=0,sticky='ew')
        
        self.imagefilew = fn = Pmw.EntryField(frame,
                                             entry_width = 35,
                                             value = "")
        fn.grid( row=row, column = 1,sticky ='w') #columnspan=2)
        self.balloon.bind(fn, "a name of tarfile that will contain frame images(.img)\nIf name is not specified-images will not be saved." )
        frame.pack( fill='x', expand=1)
        return dialog


    
    def browseFile_cb(self, filetype = "MPG"):
        """callback of the 'Save as' button of the record dialog. Opens a file browser."""

        
        fileDir = os.getcwd()
        if filetype == "MPG":
            filename = self.filename
            filenamew = self.filenamew
            filetypes = [("MPG", ".mpg")]
        elif filetype == "IMG":
            self.imagename = ""
            filename = self.imagefilename
            filenamew = self.imagefilew
            filetypes =[("Image file", ".img")] 
        else: return
        if filename:
            if os.path.exists(filename):
                fileDir = os.path.dirname(os.path.realpath(filename))
            else:
                fileDir = os.getcwd()
        file = tkFileDialog.asksaveasfilename( filetypes=filetypes,
                                               initialdir=fileDir,
                                               initialfile=None,
                                               title="Save file")
        if file:
            if filename != file:
                filename = file
                filenamew.setentry(file)


    def getFileName_cb(self):
        """ Get file name from the record dialog's entry field."""
        name = self.filenamew.get()
        if name:
            self.filename = name

    
    def getImageFileName_cb(self):
        """ Get file name from the record dialog's entry field."""
        name = self.imagefilew.get()
        if name:
            self.imagename = name

    def setCodec_cb(self, val):
        """callback of the radioselect widget of the record dialog """
        self.codecType = val

    def setCameraSize_cb(self, object=None):
        val=self.cameraSize.get()
        #print "setCameraSize:", val
        if self.cameraSizeMenub:
            self.cameraSizeMenub.configure(text = "Camera size: %s" % val)
        if val != "current":
            if not self.vf.GUI.isCameraFloating():
                self.vf.GUI.floatCamera()
            if object is not None:
                camera = object.currentCamera
            else:
                camera = self.vf.GUI.VIEWER.currentCamera
            w, h = self.resolutions[val]
            if camera.width != w or camera.height != h:
                camera.Set(width=w, height=h)
                camera.Redraw()
            camera.lift()
            


    def startRecording_cb(self, event=None):
        """Callback of the 'record' (red dot) check button """
        ifd = self.form.ifd
        cbVar = ifd.entryByName['recordB']['wcfg']['variable']
        entry = ifd.entryByName['recordB']['widget']
        recording = cbVar.get()
        self.recording = recording
        if recording:
            # activate and place the dialog just below the button that was clicked:
            self.codecw.invoke(self.codecType)
            if event:
                x = event.x_root
                y = event.y_root
                if y + self.master.winfo_height() > self.master.winfo_screenheight():
                    y = self.master.winfo_screenheight() - self.master.winfo_height()
                exitStatus = self.recordDialog.activate(geometry = '+%d+%d' % (x, y ))
            else:
                exitStatus = self.recordDialog.activate(geometry = '+%d+%d' % (100, 100 ))

            if exitStatus=='Cancel':
                self.recording = False
                ifd = self.form.ifd
                cbVar = ifd.entryByName['recordB']['wcfg']['variable']
                cbVar.set(0)
                return 
            for viActor in self.director.redrawActors.values():
                cam = viActor.object.currentCamera
                self.setCameraSize_cb(viActor.object)
                self.getFileName_cb()
                imagefilename = self.imagefilew.get()
                if imagefilename:
                    name, ext = os.path.splitext(imagefilename)
                    if ext != ".img":
                        imagefilename = name+".img"
                    self.makeImageDir(imagefilename)
                    self.recordImages = True
                else:
                    self.recordImages = False
                self.imagefilename=imagefilename
                #cam.setVideoOutputFile('movie_'+cam.uniqID+'.mpg')
                cam.setVideoOutputFile(self.filename)
                cam.setVideoParameters(outCodec=self.codecType)
                cam.nbRecordedframes = 0
                self.Play_cb()
            
                
        #print 'Recording', self.recording, self.stop


    def stopRecording_cb(self, event=None):
        """Stops camera recording. """
        if not self.recording:
            return
        self.recording = False
        for viActor in self.director.redrawActors.values():
            cam = viActor.object.currentCamera
            cam.stop()
            print 'Recorded', cam.nbRecordedframes
        ifd = self.form.ifd
        cbVar = ifd.entryByName['recordB']['wcfg']['variable']
        cbVar.set(0)
        if self.recordImages:
            self.recordImages = False
            cwd = os.getcwd()
            if os.path.exists(self.imageFolder):
                bname = os.path.basename(self.imageFolder)
                tar = tarfile.open(self.imagefilename, "w:gz")
                folder = os.path.dirname(self.imageFolder)
                os.chdir(folder)
                tar.add(bname)
                tar.close()
                os.chdir(cwd)
                shutil.rmtree(folder)
        #print 'Recording', self.recording


    def play(self, framerate=None,event=None):
        """this method is called by Play_cb()-the callback of the 'play' buttons """

        self.director.stopViewersAutoRedraw()
        self.director.dispatchEvent(PlayerStartEvent(0))
        #print 'in play() Recording', self.recording, self.stop
##         for i, viActor in enumerate(self.director.redrawActors.values()):
##             cam = viActor.object.currentCamera
##             if self.recording:
##                 #cam.setVideoOutputFile('movie_'+cam.uniqID+'.mpg')
##                 from time import strftime
##                 cam.setVideoOutputFile('movie_'+strftime("%m-%d-%Y_%H:%M")+'.mpg')
##                 cam.setVideoParameters()
##                 # don;t set recording else viewer redraw will record frames
##                 # the frames are recorded explicitely in nextFrame()
##                 #cam.videoRecordingStatus = 'recording'
##                 cam.nbRecordedframes = 0

        Player.play(self, framerate, event)

        for i, viActor in enumerate(self.director.redrawActors.values()):
            cam = viActor.object.currentCamera
            if self.recording:
                self.stopRecording_cb()

        self.director.dispatchEvent(PlayerStopEvent(-1))        
        self.director.startViewersAutoRedraw()

        if self.currentFrameIndex == self.maxFrame:
            #self.director.afterAnimationCb()
            self.GoToEnd_cb()
        elif self.currentFrameIndex == self.startFrame:
            self.GoToStart_cb()
        

    def nextFrame(self, id):
        """Overrides Player.nextFrame() method.
        Actually displays the next frame and upades the slider widget of the player
        """
        try:
            # id can be a string 'start' or 'end'
            id = int(id)
        except:
            return

        if id == self.currentFrameIndex:
            return

        if self.hasCounter and self.gui:
            self.form.ent2.delete(0,'end')
            if id < 0:
                self.form.ent2.insert(0, 'start')
            else:
                self.form.ent2.insert(0, str(id))
            if self.hasSlider:
                self.form.ifd.entryByName['slider']['widget'].set(id)
        self.currentFrameIndex = id
        if  id < 0: return
        self.director.setValuesAt(id)
        self.currentFrameIndex = id
        self.director.viewerRedraw()
        if self.recording:
            for i, viActor in enumerate(self.director.redrawActors.values()):
                cam = viActor.object.currentCamera
                imfile = None
                if self.recordImages:
                    nn = len(str(self.director.endFrame))+1
                    fstr = "%"+"0%dd.png" % (nn,)
                    imfilename = fstr%(id+1,)
                    imfile = os.path.join(self.imageFolder, imfilename)
                cam.recordFrame(force=True, imageFile=imfile)
            
        
    def update(self):
        """call in response to addMaaEvents inplemented in gui.Tk.sequenceAnimator.py """
        #print "update" 
        self.startFrame = 0
        self.maxFrame = self.endFrame = self.targetFrame = self.director.endFrame-1
        if self.hasSlider:
            sliderWidget = self.form.ifd.entryByName['slider']['widget']
            sliderWidget.configure(to=self.maxFrame+1)


    def GoToStart_cb(self, event=None):
        """Callback of '|<<'  button . Sets the player to 'start' frame."""
        #print "GoToStart_cb", self.startFrame, self.currentFrameIndex
        self.currentFrameIndex = self.startFrame-1
        self.oneDirection = 0

        if self.hasCounter and self.gui:
            self.form.ent2.delete(0,'end')
            self.form.ent2.insert(0, 'start')
            if self.hasSlider:
                self.form.ifd.entryByName['slider']['widget'].set(self.startFrame-1)


    def GoToEnd_cb(self, event=None):
        """Callback of '>>|' button. Sets the player to 'end' frame. """
        #print 'GoToEnd'
        id = self.currentFrameIndex = self.maxFrame+1
        self.oneDirection = 0

        if self.hasCounter and self.gui:
            self.form.ent2.delete(0,'end')
            self.form.ent2.insert(0, 'end')
            if self.hasSlider:
                self.form.ifd.entryByName['slider']['widget'].set(id)


    def makeImageDir(self, filename):
        import tempfile
        folder = tempfile.mktemp()
        basename = os.path.basename(filename)
        name, ext = os.path.splitext(basename)
        imageFolder = os.path.join(folder, name)
        print 'creating folder', imageFolder
        os.mkdir(folder)
        os.mkdir(imageFolder)
        #self.recordImages = True
        self.imageFolder=imageFolder


    def getImageFiles(self, filename):
        if not os.path.exists(filename):
            print "File %s does not exist" % filename
            return
        import tempfile
        basename = os.path.basename(filename)
        name, ext = os.path.splitext(basename)
        folder = tempfile.mktemp()
        tar = tarfile.open(filename)
        tar.extractall(path=folder)
        return os.path.join(folder, name)



##
## Author Michel Sanner Feb 2009
##
## Copyrights TSRI and M. Sanner
##

from mglutil.events import Event, EventHandler
from Scenario2.multipleActorsActions import MultipleActorsActions, MAAGroup

class AddMAAEvent(Event):
    """
    Event object created when a MAA is added to a target

    eventObject <- AddMAAEvent(maa)

    maa is an instance of a MultipleActorsActions. It will be available in
    the event object as the .object attribute
    """
    def __init__(self, maa):

        assert isinstance(maa, MAAGroup) or isinstance(maa, MultipleActorsActions)
        Event.__init__(self)
        self.object = maa


class RemoveMAAEvent(Event):
    """
    Event object created when a MAA is removed from the clipboard

    eventObject <- RemoveMAAEvent(maa)

    maa is an instance of a MultipleActorsActions. It will be available in
    the event object as the .object attribute
    """

    def __init__(self, maa):

        assert isinstance(maa, MAAGroup) or isinstance(maa, MultipleActorsActions)
        Event.__init__(self)
        self.object = maa


class CurrentFrameEvent(Event):
    """
    Event object created when the current frame changes in the sequence animator

    eventObject <- currentFrameEvent(frame)

    frame is a 0-based integer
    """

    def __init__(self, frame):

        Event.__init__(self)
        self.frame = frame


class PlayerStartEvent(Event):
    """
    Event object created when the player starts playing an animation

    eventObject <- PlayerStartEvent()

    frame is a 0-based integer
    """

    def __init__(self, frame):

        Event.__init__(self)
        self.frame = frame


class PlayerStopEvent(Event):
    """
    Event object created when the player stops playing an animation

    eventObject <- PlayerStopEvent()

    frame is a 0-based integer
    """

    def __init__(self, frame):

        Event.__init__(self)
        self.frame = frame



##
##  Author Michel F. Sanner May 2007
##

from time import sleep
import warnings, weakref, numpy

from actor import Actor
from Scenario2 import MaxKF

try:
    from DejaVu.scenarioInterface.animations import RedrawActor
    foundDejaVu = True
except ImportError:
    foundDejaVu = False
    
## from SimPy.Simulation import Process, initialize, activate, simulate, now, hold
from Scenario2.multipleActorsActions import MultipleActorsActions, MAAGroup, ActorProcess


## class Director(MultipleActorsActions):
##     """
## """
##     def __init__(self):

##         MultipleActorsActions.__init__(self)

##         self.currentFrame = 0
##         self.endFrame = 0

##         self.gui = None # will be set when a gui is created
##         self.recordingActors = []  # list of actors with recording flag on
##            # i.e. actor.recording!='stopped'
##         # FIXME should have a set and delete method to check type etc
##         self.afterAnimation = [] # list of functions to be called after
##                                  # play ends
##         self.scenarios = {}
##         self.autotrackDict = {}
##         self.redrawActor = None
##         self.moveForward = True
##         self.maxFrame = 0
##         self.needsRedraw = False
        

##     def addScenario(self, name, scenario):
##         self.scenarios[name] = scenario


##     def start(self, setDirector = True, title = None):
##         if self.gui is None:
##             gui = DirectorGUI(self, title = title)
##             gui.root.protocol('WM_DELETE_WINDOW', gui.root.withdraw)
##             self.gui = gui
##             if setDirector:
##                 for scenario in self.scenarios.values():
##                     scenario.setDirector()
##         else:
##             if title:
##                 oldtitle = self.gui.root.title()
##                 if oldtitle != title:
##                    self.gui.root.title(title) 
##             self.gui.root.deiconify()


## ##     def deleteActor(self, actor):
## ##         # delete this actor and all its actions
## ##         name = actor.name
## ##         if self.gui:
## ##             self.gui.deleteActors()
## ##         if actor in self.actors:
## ##             self.actors.remove(actor)
## ##         if hasattr(actor, "guiVar"):
## ##             if actor.guiVar:
## ##                 actor.guiVar.set(0)
## ##         #if hasattr(actor.object, "animatedProperties"):
## ##         #    del actor.object.animatedProperties[name]
## ##         if hasattr(actor.scenario, "animatedProperties"):
## ##             del actor.scenario.animatedProperties[name]
## ##         if self.gui:
## ##             if self.gui.selectedFrames.has_key(name):
## ##                 self.gui.selectedFrames.pop(name)
## ##             if self.gui.autotracking:
## ##                 if self.autotrackDict.has_key(name):
## ##                     self.autotrackDict.pop(name)
## ##                     # this will update the dictioanary with a new "clean"
## ##                     # version of the deleted actor.
## ##                     self.autotrackDict.update(actor.scenario.getNewActors())
                
## ##             self.gui.drawActors()
## ##         ####################FIX THIS ###########
## ##         if len(actor.kfSetId):
## ##             for kf, setId in actor.kfSetId.items():
## ##                 kfset = self.gui.kfSets.get(setId)
## ##                 if kfset:
## ##                     kfset.removeKeyFrame(name, kf)
## ##                     if kfset.getNumberOfFrames() == 1:
## ##                         # remove the set
## ##                         self.gui.removeKFSet(kfset)

        

##     def getLastFrame(self):
##         # returns number of frames in the simulation
##         return self.endFrame

    
##     def getLastFrameWithChange(self, actors = None):
##         # returns the last frame in which a value changes
##         if not actors:
##             actors = self.actors
##         last = None
##         for actor in actors:
##             #if actor.isFullRange(): continue
##             if actor.visible:
##                 lastFrame = actor.getLastKeyFrame()
##                 if last is None:
##                     last = lastFrame
##                 else:
##                     if lastFrame.pos > last.pos:
##                         last = lastFrame
##         return last


##     def updateEndFrame(self, end, updateGUI=1):
##         if not end:
##             return
        
##         if end > self.endFrame:
##             self.endFrame = end
##             if self.gui and updateGUI:
##                 self.gui.setDuration(end)

##         else:
##             last = self.getLastFrameWithChange()
##             if last is not None:
##                 last = last.pos
##             else:
##                 last = 0
##             if end < last+10 : # when recording we want to make sure there are frame to the right
##                 self.endFrame = last +10
##             else:
##                 self.endFrame = end
##             if self.gui and updateGUI:
##                 self.gui.setDuration(self.endFrame)


##     def run(self, forward=True):
##         if len(self.actors)==0:
##             #print "no actors"
##             return
##         initialize()
##         redraw = self.redrawActor
##         start = 0
##         #end = self.endFrame
##         self.maxFrame = 0
##         maxFrame = self.getLastFrameWithChange()
##         if maxFrame is not None:
##             self.maxFrame = maxFrame.pos

##         end = self.maxFrame
##         gui = self.gui
##         if gui:
##             if not (gui.startFrame == 0 and gui.stopFrame == 1):
##                 start = gui.startFrame
##                 end = gui.stopFrame
##                 self.maxFrame = end + start
##         self.moveForward = forward
##         if redraw:
##             redraw.object.stopAutoRedraw()
##             proc = RedrawActorProcess(redraw)
##             activate(proc, proc.execute(), at=start, prior=True)
##             #print 'activated redraw', 
##         for actor in self.actors:
##             if actor.name != "redraw":
##                 proc = ActorProcess(actor)
##                 activate(proc, proc.execute(), at=start, prior=True)
##                 #print 'activated ', actor.name, action
##         #print 'simulate', self.endFrame
##         simulate(until=end)
##         #self.currentFrame = self.endFrame
##         if forward:
##             #self.currentFrame = self.endFrame
##             self.currentFrame = end
##         else:
##            self.currentFrame = start
##         # call callback after animation completes
##         for f in self.afterAnimation:
##             f()
##         if self.redrawActor:
##             vi = self.redrawActor.object
##             vi.startAutoRedraw()


##     def setValuesAt(self, frame, actor = None):

##         if actor:
##             actors = [actor]
##         else:
##             actors = self.actors

##         for actor in actors:
##             actions = actor.actions
##             if actor.name != "redraw" and  actor.playing:
##                 if len(actor.valueGenerators):
##                     value = actions.getValueAt(frame)

##                     if value != 'Nothing There':
## #                        if redraw:
## #                            self.needsRedraw = True
##                         actor.setValue(value)
                                
## #        if redraw:
## #            redraw.setValue()

##     def gotoFrame(self, frame, set=True):
##         if frame<0:
##             frame = 0
##         elif frame>self.endFrame:
##             frame = self.endFrame
##         if frame==self.currentFrame:
##             return
##         self.currentFrame = frame
##         if set:
##             self.setValuesAt(frame)
##         if self.gui:
##             self.gui.placeTimeCursor(frame)
            

## ##     def createActorFromData(self, obj, propname, scenarioInterface,filename, fields,
## ##                             data, start = 0, end=-1):
## ##         # Create a special actor whose valueGenerator returns values from a data 
## ##         # sequence such as list ar numpy array, This data is read from a file (.adat)->
## ##         # see actor.py/adatFileParser

## ##         # obj - animatable object
## ##         # propname - the object's animated property name
## ##         # scenarioInterface - scenario interface that is n self.scenarios and "knows" how
## ##         # to create a regular actor for this object (i.e has createActor() method).
## ##         # start - first keyframe of the actor
## ##         # end - last keyframe of the actor ( if end == -1, then the number of frames between
## ##         # first and last keyframes is len(data)-1, or one data step per each keyframe).

## ##         if data is None:
## ##             print "createActorFromData: data is None"
## ##             return

## ##         assert scenarioInterface in self.scenarios.values()
## ##         actor = None
## ##         if data is not None:
## ##             assert type(data).__name__ in ('ndarray', 'list')
## ##             nsteps = len(data)
## ##             firstVal = data[0]
## ##             lastVal =  data[nsteps-1]
## ##             #print filename, len(data), data[0], propname
## ##             # create an actor with ReadDataInterpolator valuegenerator
## ##             #try:

## ##             # this will create file actor with two keyframes : "0" and "nsteps-1" 
## ##             actor = scenarioInterface.createActor(obj, propname, check=False,
## ##                                                   addToDirector=False,
## ##                                                   actorData = (filename, data))
## ##             actor.datafields = fields
## ##             #except:
## ##             #    warnings.warn("Failed to create file actor %s" % propname)
## ##             #    return None
## ##             if actor:
## ##                 self.addActor(actor)
## ##                 #print "start=", start, "end=", end, "nsteps=", nsteps
## ##                 if end == -1:
## ##                     end = start + nsteps - 1
## ##                 if end != nsteps-1:
## ##                     if self.gui:
## ##                         #this will move "nsteps-1" keyframe  to "end" keyframe 
## ##                         self.gui.selectKeyFrame_cb(actor, nsteps-1, newframe=end)
## ##                 if start != 0:
## ##                     if self.gui:
## ##                         # this will move "0" keyframe to "start" keyframe
## ##                         self.gui.selectKeyFrame_cb(actor,0,newframe=start) 
                    
## ##         return actor
            

from mglutil.events import EventHandler
from Scenario2.events import AddMAAEvent, RemoveMAAEvent, CurrentFrameEvent,\
     PlayerStartEvent, PlayerStopEvent

from DejaVu.states import setRendering, setOrientation

class MAADirector(EventHandler):
    """
    an MAADirector object contains a list of MultipleActorsActions (MAA) objects
    and their corresponding positions, i.e. absolute frame number in the
    animation.

    the (maas, position) tuples are stored in the .maas attribute

    attributes:

    .currentFrame: current frame in the animation
    .maxFrame:
    .endFrame: frame after the last frame for which there is a value
    
    Methods:

    bool <- addMAAat(self, maa, position)
    maa, index <- removeMAA(maa, position)
    None <- play(self, fromPos=0, toPos=None, forward=True, callback=None)
    None <- def setValuesAt(self, frame)
    string <- getMAASourceCode(self, variable, indent=0, showwarning=True)
    """

    def __init__(self):

        EventHandler.__init__(self)
        self.maas = [] # will hold the list of [MAA, position] pairs
        self.currentFrame = 0
        self.endFrame = 0  # frame after the last frame for which there is a value
        self.maxFrame = 0

        # these are used internally
        self.moveForward = True # truwe when palying forward
        self.needsRedraw = False
        self.redrawActors = {} # {Viewer Object: RedrawActor}


    def addMAAat(self, maa, position, check=True):
        """
        adds an MAA object to be played at a given position.

        bool <- addMAAat(maa, position)
        
        the MAA can only be added if none of its actors overlaps with the same
        actor in an existing MAA. Returns True is maa was added.

        Event:
            generates a AddMAA event
        """
        if not isinstance(maa, MultipleActorsActions) and not isinstance(maa, MAAGroup):
            raise TypeError, "in addMaaat() maa must be an instance of MultipleActorsActions or MAAGroup"
        #assert isinstance(maa, MultipleActorsActions)
        assert isinstance(position, int)

        maxiAll = 0
        # check that no actor in maa overlaps with the same actor
        redrawInd = []
        for i, actor in enumerate(maa.actors):

            actions = actor.actions
            mini, maxi = actions.boundingFrames(position)
            if maxi < MaxKF and maxi > maxiAll:
                maxiAll = maxi
            if check:
                for maa1, pos1 in self.maas:
                    #if actor.name[:6]=='redraw':
                    #    self.redrawActors[actor.object] = actor
                    #    continue
                    # does actor exist in maa1?
                    actor1 = maa1.findActor(actor)
                    if actor1 is None: # actor not in maa1
                        continue
                    if foundDejaVu and isinstance(actor1, RedrawActor):
                        continue
                    # does actions at position overlap with actor1 at pos1
                    mini1, maxi1 = actor1.actions.boundingFrames(pos1)
                    if maxi<mini1: continue
                    elif mini>maxi1: continue
                    else: return False
                    if mini>maxi1: continue
                    if maxi<mini1: continue
                    else: return False

            # if the maa contains a redraw actor add it to self.redrawActors
            # and remove it from the maa (after the "for" loop)
            if foundDejaVu and isinstance(actor, RedrawActor):
                if not self.redrawActors.has_key(actor.object):
                    self.redrawActors[actor.object] = actor
                    redrawInd.append(i)
                    
        for i in redrawInd:            
            maa.actors.pop(i)
                    
        # update self.endFrame
        if maxiAll > self.endFrame:
            self.endFrame = maxiAll + 1

        #add maa
        self.maas.append( [maa, position] )
        if not hasattr(maa, "startFlag"):
            maa.startFlag = "after previous"
        maa.startOffset = 0
        maa._director = weakref.ref(self)
        if isinstance(maa, MAAGroup):
            for m in maa.maas:
                m._director = weakref.ref(self)
        # create and dispatch event
        self.dispatchEvent(AddMAAEvent(maa))
        
        return True


    def stopViewersAutoRedraw(self):
        """
        Stop autoRedraw in all Viewer associates with RedrawActors
        """
        for actor in self.redrawActors.values():
            #print "stopped autoredraw"
            actor.object.stopAutoRedraw()
            
    def startViewersAutoRedraw(self):
        """
        Start autoRedraw in all Viewer associates with RedrawActors
        """
        for actor in self.redrawActors.values():
            #print "start autoredraw"
            actor.object.startAutoRedraw()

    def viewerRedraw(self):
        """
        make sure all Viewers will redraw
        """
        #print 'Viewer Redraw'
        for actor in self.redrawActors.values():
            if self.needsRedraw and actor.playing:
                actor.setValue()
                self.needsRedraw = False
                actor._maa().needsRedraw = False

        
    def setOrientation(self, orient):
        """
        make sure all Viewers will redraw
        """
        
        for actor in self.redrawActors.values():
            if actor.playing:
                vi = actor.object
                root = vi.rootObject
                if orient.has_key(root.fullName): # snapshot orient 
                    allkeys = orient.keys()
                    ort = {}
                    ort.update(orient[root.fullName])
                    allkeys.remove(root.fullName)
                    if orient.has_key('fieldOfView'):
                        ort['fieldOfView'] = orient['fieldOfView']
                        allkeys.remove('fieldOfView')
                    if orient.has_key('lookFrom'):
                        ort['lookFrom'] = orient['lookFrom']
                        allkeys.remove('lookFrom')                                  
                    setOrientation(root, ort)
                    for name in allkeys:
                        obj = vi.FindObjectByName(name)
                        if obj.fullName == name:
                           setOrientation(obj, orient[name])
                else:
                    setOrientation(actor.object.rootObject, orient)

        
    def setRendering(self, rendering):
        """
        make sure all Viewers will redraw
        """
        for actor in self.redrawActors.values():
            if actor.playing:
                setRendering( actor.object, rendering )

        
    def play(self, fromPos=0, toPos=None, forward=True, callback=None):
        """
        play through the list of MAAs

        None <- play(self, fromPos=0, toPos=None, forward=True, callback=None)

        fromPos: starting frame for play back
        toPos: ending frame for play back
        forward: True when playing direction is forward,
                 False when playing direction is backward
        
        """
        ## SEEMS NOT USED
        self.dispatchEvent(PlayerStartEvent(fromPos))

        if toPos is None:
            toPos = self.endFrame-1

        if len(self.maas)==0:
            return

        self.needsRedraw = False

        self.stopViewersAutoRedraw()
        
        for i in xrange(fromPos, toPos+1):
            self.setValuesAt(i)
            if callback:
                callback(i)
            self.viewerRedraw()
            
        self.startViewersAutoRedraw()

        #self.afterAnimationCb()

        self.dispatchEvent(PlayerStopEvent(i))


    def afterAnimationCb(self):
        """
        Call callback for each MAA to restore the state of objects drive by MAAs
        """
        for maa, off in self.maas:
            maa.afterAnimation_cb()
            

    def setValuesAt(self, frame):
        """
        Goto to a specified frame and issues values for all MAAs

        None <- def setValuesAt(self, frame):

        # FIXME this should also issue ending values of MAAs that ended before
        """

        # create and dispatch event for current frame change
        self.dispatchEvent(CurrentFrameEvent(frame))
        #stopAutoRedraw = False
        needsRedraw = False
        for maa, pos in self.maas:
            #print 'setting frame:', frame, maa.name, pos + maa.firstPosition
            if frame == pos + maa.firstPosition: # first Frame
                firstFrame = True
                if hasattr( maa, 'orient') and maa.forceOrient:
                    if not maa.setOrient:
                        self.setOrientation(maa.orient)
                        
                if hasattr(maa, "geomOrients") and maa.setOrient:
                        #self.stopViewersAutoRedraw()
                        #stopAutoRedraw = True
                        for o, val in maa.geomOrients.items():
                            setOrientation(o, val)
                
                if hasattr( maa, 'rendering') and maa.forceRendering:
                    self.setRendering(maa.rendering)
            else:
                firstFrame = False

            if frame >= pos+ maa.firstPosition and \
                   frame <= pos + maa.lastPosition:
                #print "setting values at frame %d , pos: %d for maa " % (
                #        frame,pos), maa.name

                maa.setValuesAt(frame, pos)
##                 if stopAutoRedraw:
##                     self.startViewersAutoRedraw()
##                     stopAutoRedraw = False
                if maa.needsRedraw: needsRedraw = True
        self.needsRedraw = needsRedraw
        

            #elif frame > maa.lastPosition+off:
            #    maa.setValuesAt(maa.lastPosition+off, off)



    def removeMAA(self, maa, position):
        """
        remove maa from MAADirector.

        maa, index <- removeMAA(maa, position)
        
        arguments:
           maa: instance of MAA;
           position: the specified MAA starting position in the MAADirector

        return value:
           returns the removed maa and its index in the director's MAA list

        Event:
            generates a RemoveMAA event

        Exceptions:
            raises a RuntimeError if MAA is not found in self.maas
        """

        # get index if it was not specified
        index = 0
        _maa = None
        for maa1, pos in self.maas:
            if maa==maa1 and pos==position:
                _maa = maa1
                break
            index += 1

        if _maa is None:
            raise RuntimeError("MAA not found")

        # remove maa
        rmaa = self.maas.pop(index)
        
        # check if self.redrawActors.values() contains a redraw actor from the removed maa:
        for obj,redrawactor in self.redrawActors.items():
            if redrawactor._maa()==maa:
                del self.redrawActors[redrawactor.object]
##         redrawactor = None 
##         for actor in rmaa[0].actors:
##             if foundDejaVu and isinstance(actor, RedrawActor):
##                 redrawactor = actor
##                 break
##         if redrawactor in self.redrawActors.values():
##             # remove the redraw actor from self.redrawActors dict, check if the rest of maas
##             # contain a redraw for the same object:
##             del (self.redrawActors[redrawactor.object])
##             obj = redrawactor.object
##             for maa in self.maas:
##                 if self.redrawActors.get(obj):
##                     break
##                 for actor in maa:
##                     if foundDejaVu and isinstance(actor, RedrawActor):
##                         if actor.object == obj:
##                             self.redrawActors[actor.object] = actor
##                             break
                    
        # create and dispatch event
        self.dispatchEvent(RemoveMAAEvent(rmaa[0]))

        # update director's endFrame
        self.updateEndFrame()

        return maa1, index


    def updateEndFrame(self):
        # update director's endFrame
        maxi = 0
        if len(self.maas)==0:
            self.endFrame = maxi
            return
        
        for maa, position in self.maas:
            end = position + maa.lastPosition
            if end > maxi:
                maxi = end

        self.endFrame = maxi + 1


    def computePosition(self, maa, startFlag=None, startOffset=None):
        """
        Compute the absolute starting frame of the specified maa
        
        position, index <- computePosition(maa, startFlag=None, startOffset=None)
        arguments:
           maa      - compute new position for this maa
           startFlag - can be 'with previous' or 'after previous'
           startOffset - start offset from the position specified by startFlag

        return values:
           position - absolute starting frame in animation
           index    - index of the specified maa in the self.maas list.
        """
        index = 0
        _maa = None
        _pos = None
        for maa1, pos in self.maas:
            if maa==maa1:
                _maa = maa1
                _pos = pos
                break
            index += 1

        if _maa is None:
            raise RuntimeError("MAA not found")

        if startFlag is None:
            startFlag = maa.startFlag
        if startOffset is None:
            startOffset = maa.startOffset

        if index == 0:
            position = startOffset
        else:
            # find previous maa in the list of self.maas
            prevMaa, prevPos = self.maas[index-1]
            # compute new position
            mini, maxi = prevMaa.boundingFrames(position = prevPos)
            if startFlag == "after previous":
                position = maxi+ 1 + startOffset
                #if startOffset == 0 :
                #    position = position + 1
            elif startFlag == "with previous":
                position = mini + startOffset
            else:
                raise RuntimeError("Wrong starting position: %s" % startFlag)
        return (position, index)


    def getMAASourceCode(self, variable, indent = 0, showwarning = True):
        """
        Return python code creating the list of maas

        string <- getMAASourceCode(self, variable, indent=0, showwarning=True)

        """

        tabs = " "*indent
        newtabs = tabs + 4*" "
        # define the showwarning variable that is used in the source code returned by
        # maa.getSourceCode()
        lines = tabs + """showwarning = %s\n""" % showwarning
        if not len (self.maas): return ""
        i = 0
        for maa in self.maas:
            if hasattr(maa[0], "getSourceCode"):
                maasrc = maa[0].getSourceCode("maa%d"%i)
                if len(maasrc):
                    lines += tabs + """maa%d = None\n"""%i
                    lines += tabs + maasrc + """\n"""
                    lines += tabs + """if maa%d is not None and len(maa%d.actors)> 0: \n""" % (i, i)
                    lines += newtabs + """%s.addMAAat(maa%d, offset+%d, check=False)\n""" % (variable, i,maa[1])
                    i = i+1
        return lines
            

##     def checkPosition(self, maa, position):
##         """check that no actor in the maa (placed at the specified position)
##         will overlap with the same actor of the maas in the list """
##         for maa1, pos1 in self.maas:
##             if maa1 == maa: continue
##             for actor in maa.actors:
##                 if actor.name[:6] == 'redraw':
##                     continue
##                 actor1 = maa1.findActor(actor)
##                 if actor1 is None: # actor not in maa1
##                     continue
##                 #same actor found
##                 mini, maxi = actor.actions.boundingFrames(position = position)
##                 mini1, maxi1 = actor1.actions.boundingFrames(pos1)
##                 if maxi<mini1: continue
##                 elif mini>maxi1: continue
##                 else:
##                     position = self.checkPosition(maa, maxi1 + 1)
##         return position
                    
                        
if __name__=='__main__':
    from datatypes import FloatType
    from interpolators import FloatScalarInterpolator

    class foo:

        def __init__(self):
            self.a= 2.0

    f = foo()

##     # create an actor
##     actor = Actor('test', f, 0.5, FloatType, FloatScalarInterpolator)
##     assert len(actor.keyframes.values) == 1
##     assert actor.keyframes.values[0] == 0.5

##     # create a director
##     d = Director()
##     d.addActor(actor)
    
##     # run the animation
##     print "Run keyframe 0"
##     d.run()

##     # add a keyframe
##     actor.setKeyframe( 10, 10.5)
##     print "Run keyframe 0 and 1"
##     d.run()

##     # add a keyframe at 15
##     actor.setKeyframe( 20, 10.5)
##     actor.setKeyframe( 25, 5.5)
##     assert len(actor.valueGenerators)==4
##     assert len(actor.keyframes.kfs)==4
##     print "Run keyframe 0,1,2,3"
##     d.run()

##     # stop interpolation between keyframe 1 and 2 (i.e interval 1)
##     actor.valueGenerators[1].configure(active = False)
##     #actor.setValueGenerator( None, segment=1)
##     print "Run keyframe 0,1 and 2,3"
##     d.run()
    



##
## Author Michel Sanner Feb 2009
##
## Copyrights TSRI and M. Sanner
##

from mglutil.events import Event, EventHandler

from Scenario2.multipleActorsActions import MultipleActorsActions


class AddMAAToClipBoardEvent(Event):
    """
    Event object created when a MAA is added to the clipboard

    eventObject <- AddMAAToClipBoardEvent(maa)

    maa is an instance of a MultipleActorsActions. It will be available in
    the event object as the .object attribute
    """
    def __init__(self, maa):

        assert isinstance(maa, MultipleActorsActions)
        Event.__init__(self)
        self.object = maa


class RemoveMAAFromClipBoardEvent(Event):
    """
    Event object created when a MAA is removed from the clipboard

    eventObject <- RemoveMAAFromClipBoardEvent(maa)

    maa is an instance of a MultipleActorsActions. It will be available in
    the event object as the .object attribute
    """

    def __init__(self, maa):

        assert isinstance(maa, MultipleActorsActions)
        Event.__init__(self)
        self.object = maa



class Clipboard(EventHandler):
    """
    Scenario clipboard class.

    This class holds a list of MultipleActorActions objects
    """

    def __init__(self):
        """
        Clipboard object constructor

        clipbardObject <- Clipboard()
        """
        EventHandler.__init__(self)
        self.maas = []


    def addMaa(self, maa, checkForDuplicates=True):
        """
        add a MultipleActorsActions (MAA) to this clipboard

        Bool <- Clipboard.addMAA(maa, checkForDuplicates=True)

        maa is an instance of a MultipleActorsActions
        if checkForDuplicates is True the same maa cannot be added twice
        
        Returns True if the MAA is added succefully and False else.
        A MAA cannot be added if this MAA is already on the Clipboard

        if the MAA is add an AddMAAToClipBoard is created and dispatched
        """
        assert isinstance(maa, MultipleActorsActions)

        # check if maa can be added
        if checkForDuplicates and maa in self.maas:
            return False

        # add maa
        self.maas.append(maa)

        # create and dispatch event
        self.dispatchEvent(AddMAAToClipBoardEvent(maa))

        return True


    def removeMaa(self, maa):
        """
        remove a MultipleActorsActions (MAA) from this clipboard

        Bool <- Clipboard.removeMaa(maa)

        maa is an instance of a MultipleActorsActions

        Returns True if the MAA is removed succefully or False 
        if the MAA is not found

        if the MAA is add an RemoveMAAToClipBoard is created and dispatched
        """
        assert isinstance(maa, MultipleActorsActions)

        # check if maa can be removed
        if maa not in self.maas:
            return False
        
        # remove maa
        self.maas.remove(maa)
        
        # create and dispatch event
        self.dispatchEvent(RemoveMAAFromClipBoardEvent(maa))

        return True





##
##  Author Michel F. Sanner Jan 2009
##
import weakref
from Scenario2.keyframes import KF, Interval
from Scenario2 import MaxKF
from Scenario2.keyframes import AutoCurrentValueObject, KFValueFromFunction

class Actions:
    """
    container for a list of keyframes and intervals

    attributes:
    .keyframes: list of chronologically sorted Scenario2.keyframes.KF objects
    .intervals: list of chronologically sorted Scenario2.keyframes.Interval obj

    methods:
    KF <- addKeyframe(self, kf)
    None <- deleteKeyframe(self, kf)

    Bool <- actions.addActionsAt(srcActions, position, check=True)

    Bool <- addInterval(self, interval):

    Interval <- getInterval(self, position)

    value <- getValue(self, frame)

    Actions <- copyAt(self, position)
    
    firstFrame, lastFrame <- boundingFrames(self, position=0)
    """

    def __init__(self, actor=None):
        if actor:
            self.actor = weakref.ref(actor)
        else:
            self.actor = None
        self.keyframes = [] # sorted chronologically
        self.intervals = [] #these have to be sorted chronologically


    def getStringRepr(self, pos1=0, pos2=None, actorName='ACTOR______NAME'):
        """Return a string description of action. The description contains
        a list of position labels, containing all the mutiple of 10 positions,
        a string representing the time line and a string representing the
        actions. The full string is a multi line strign that will display all
        the information. If provided, the name will appear in front of the
        actStr and use 15 charaters

        posLabel, posStr, actStr, fullStr <-getStringRepr(pos1=0, pos2=None)

        if pos2 is None pos2 will be the position of the last keyframe

        posLabel is a list of all multiple of 10 positions (i.e. 'I' in poStr)
        
        The posStr has an '|' at all multiples of 10 and a ":" the multiple of 5
        that are not multiple of 10 (e.g. "   |    :    |     : ")

        The actStr string contains:
          ' ': a blank space for each position where no value
          'x': where there is a keyframe
          '-': where there is an active interval
          '.': where there is an inactive interval
          '|' every position that is a multiple of 10

          e.g. 'x---------x    x    x---------x'

        FullString example:
        
                             1         2         3
                             0         0         0         0
                             |    :    |    :    |    :    |
             ACTOR______NAME x.........x    x    x---------x

             interval (0,10) is inactive, there is an isolated keyframe at 15
             and an active interval from 20 to 30
        """

        nameLength = 30
        posStr = " "*(nameLength+1)
        sformat = "%0"+"%ds"%nameLength
        actStr = sformat%actorName[:nameLength]
        posLabels = []

        if len(self.keyframes)==0:
            return [], posStr, actStr, actStr+"No Keyframes"

        if pos2 is None:
            pos2 = self.keyframes[-1].pos

        nblines = len(str(pos2)) # how many lines we need for labels
        labelLines = []
        for i in range(nblines):
            labelLines.append( " "*(nameLength+1) )

        for i, pos in enumerate(xrange(pos1,pos2+1)):
            obj = self.getInterval(pos)

            # build posStr
            rpos = i+pos1
            if rpos%10==0 or i==0:
                if (rpos%10==0):
                    posStr += '|'
                else:
                    posStr += ' '
                posLabels.append(rpos)
            elif rpos%5==0:
                posStr += ':'
            else:
                posStr += ' '

            # build label lines 
            if (rpos)%10==0 or i==0:
                lab = str(rpos)
                nbl = len(lab)
                for j in range(0,nblines-nbl):
                    labelLines[j] += ' '
                for j,k in enumerate(range(nblines-nbl,nblines)):
                     labelLines[k] += lab[j]
            else:
                for j in range(nblines):
                    labelLines[j] += ' '

            # build actStr
            if obj is None:
                actStr += ' '

            elif isinstance(obj, KF):
                actStr += 'x'

            elif obj.active: # we are on an active interval
##                 if (rpos%10==0):
##                     actStr += '+'
##                 elif rpos%5==0:
##                     actStr += ':'
##                 else:
##                     actStr += '-'
                actStr += '-'
                
            else: # we are on an inactive interval
                actStr += '.'

        # build fullStr
        fullStr = """"""
        for i in range( nblines):
            fullStr += labelLines[i]+'\n'
        fullStr += posStr+'\n'
        fullStr += actStr
        
        return posLabels, posStr, actStr, fullStr

    
    def addKeyframe(self, kf):
        """
        add a  keyframe kf.
        Kf can be a KF object or a (position, value) pair.
        this method returns and existing KF or the one we sent into the method
        """
        if not isinstance(kf, KF):
            kf = KF( *kf )
        
        # find what is on
        obj = self.getInterval(kf.pos)
        if obj is None: # there is nothing at this position
                        # we simply add a copy of this keyframe
            self._insertKeyframe(kf)
            return kf
        
        elif isinstance(obj, KF):
            if kf.isLinked():
                return  None# a linked keyframe cannot be added on top of a keyframe

            elif kf.__class__ != obj.__class__: # we replace the keyframe
                kf.leftInterval = obj.leftInterval
                if kf.leftInterval is not None:
                    kf.leftInterval.kf2 = kf
                kf.rightInterval = obj.rightInterval
                if kf.rightInterval is not None:
                    kf.rightInterval.kf1 = kf
                self.keyframes.remove(obj)
                self._insertKeyframe(kf)
                return kf

            elif kf.valueObject.__class__ == obj.valueObject.__class__:
                # we over write the value
                obj.setValue(kf.getValue())
                return obj

            else:
                # we need to overwrite KF's value object
                obj.valueObject = kf.valueObject
                return obj

        elif isinstance(obj, Interval):
            # put kf in list of keyframes
            self._insertKeyframe(kf)

            if obj.isLinked(): # this is a linked interval
                off = kf.pos - obj.k1.pos
                self.splitIntervalWithKF(obj, kf)
                newInter = kf.rightInterval


                for inter in linkedInter[1:]:
                    il = newInter.linkedCopy(inter.k1.pos+off)
                    il.setKF2( inter.k2 )
                    kf1 = il.kf1
                    self._insertKeyframe(kf1)
                    inter.setKf2( kf1 )

            else: # not linked
                self.splitIntervalWithKF(obj, kf)

            return kf

    def deleteKeyframe(self, kf):
        """
        Delete keyframe kf if kf in not in any interval
        """
        try:
            for i in self.intervals:
                assert i.kf1!=kf
                assert i.kf2!=kf
        except AssertionError:
            print "could not remove kf, belongs to interval", kf, kf.pos
            return
        self.keyframes.remove(kf)


        
    def splitIntervalWithKF(self, inter, kf):
        """
        Split the interval inter using keyframe kf.
        kf is known to be in the interval. 
        """
        newInter = Interval(kf, inter.kf2)

        if inter.valGen.generator:
            gen = inter.valGen.generator
            generator = gen.clone()
            newInter.setValGenGenerator(generator)

        # set left and right for kf
        kf.leftInterval = inter
        kf.rightInterval = newInter

        # set kf as last of old interval
        inter.setKF2(kf)

        self._insertInterval(newInter)

##         for i,inter in enumerate(self.intervals):
##             if inter==obj:
##                 break
##         self.intervals.insert(i+1, newInter)

     
    def _insertInterval(self, inter):
        """
        insert the interval inter into self.intervals respecting the chronolgy. 
        The method does not perform any checks. It is assumed that inter.kf1 and
        inter.kf2 are in self.keyframes
        """
        
        p1 = inter.kf1.pos
        n=0 # needed when len(self.intervals)==0
        for n,i in enumerate(self.intervals):
            if i.kf1.pos > p1:
                break
        if len(self.intervals):
            if p1 > i.kf1.pos: # this happens when we split last interval or add a new interval after the last one
                n = n+1
        self.intervals.insert(n, inter)
        return


    def _insertKeyframe(self, kf):
        """
        insert the keyframe fk into self.keyframes respecting the chronolgy
        """
        if len(self.keyframes)==0:
            self.keyframes.append(kf)
            return
        
        for i,k in enumerate(self.keyframes):
            if k.pos==kf.pos:
                raise ValueError("position %d already has a keyframe"%k.pos)
            if k.pos > kf.pos:
                self.keyframes.insert(i, kf)
                return
        self.keyframes.append(kf)


    def boundingFrames(self, position=0):
        """
        return first and last frame of an action object
        """
        first = 0
        last = MaxKF
        if len(self.keyframes)>0:
            first = self.keyframes[0].pos+position
            last = self.keyframes[-1].pos+position
        return first, last


    def isLinked(self):
        """
        Return True if any object in the action is linked
        """
        for k in self.keyframes:
            if k.isLinked():
                return True

        for i in self.intervals:
            if i.isLinked():
                return True

        return False


    def keyframeBetween(self, pos1, pos2):
        """
        return true if there is a keyframe between the 2 positions
        """
        for k in self.keyframes:
            if k.pos >= pos1:
                if k.pos <= pos2:
                    return True
                else:
                    return False
                

    def checkAddLinkedActionsAt(self, action, position):
        """
        Determine whether a linked copy of the action can be added
        to this action object at the specified position.
        The addition of the action is allowed only if it does not overlap any
        KF or interval in current action (self) when placed at the specified position.
        """

        if len(actions.keyframes)==0:
            return False
        
        firstPos = actions.keyFrames[0].pos
        #delta = position - firstPos
        delta = position

        # get first and last frame for the action to be added 
        mini, maxi = action.boundingFrames()
        mini += delta
        maxi += delta

        # check that no keyframe of self is withing mini and maxi
        for k in self.keyframes:
            if k.pos >= mini and k.pos <= maxi:
                return False

        # now we know that no KF is within mini and maxi
        # we need to check that no interval of self spans the whole action object
        # this can only happen if there is an interval i in self for which the first
        # KF is BEFORE mini and the last KF is AFTER maxi
        for i in self.intervals:
            k1 = i.kf1
            k2 = i.kf2
            if k1.pos < mini and k.pos > maxi:
                return False

        return True

    
    def checkAddActionsAt(self, actions, position):
        """
        Determine whether the physical copy of the keyframes and intervals in action
        can be added to this action object at the specified position.
        The following rules are enforced:
           1. no KF to be added (at the position) falls onto an interval
           2. no inside of an interval to be added overlaps with either a KF or an Interval
        """
        if len(actions.keyframes)==0:
            return
        
        firstPos = actions.keyframes[0].pos
        #delta = position - firstPos
        delta = position

        # no KF to be added falls onto an interval
        for k in actions.keyframes: # loop over KFs kf to be added
            # check what is in self at position kf.pos
            obj = self.getInterval(k.pos+delta)
            # if it is an Interval we cannot paste actions
            if isinstance(obj, Interval):
                return False

        # no KF from actions to be added is inside of an interval
        # hence the each KF to be added either falls on nothing or a KF
        # in self

        # added overlaps with either a KF or an Interval
        for i in actions.intervals: # loop over interals to be added

            # check that self has no KF within interval to be added i
            if self.keyframeBetween(i.kf1.pos+1+delta, i.kf2.pos-1+delta):
                return False
                
        return True

            
    def addLinkedActionsAt(self, actions, position, check=True):
        """
        Creates a linked copy of the actions at specified position
        and adds it to this actions object.
        If check is True, the method first checks if linked actions can be added
        at the position.
        """
        if check:
            if not self.checkAddLinkedActionsAt(actions, position):
                return

        actions1 = actions.linkedCopyAt(position)

        for sourcekf in actions1.keyframes:
            destkf = self._insertKeyframe( sourcekf )
            
        # insert interval
        for inter in actions1.intervals:
            self._insertInterval(inter)


    def addActionsAt(self, srcActions, position, check=True):
        """
        Add srcActions at a given position to this actions object

        Bool <- actions.addActionsAt(srcActions, position, check=True)
        
        Creates a physical copy of the actions at specified position
        and adds it to this actions object.
        
        If check is True, the method first checks if actions can be added
        at the position adn returns False if it can;t be added. If the
        action is added successfully True is returned.
        """
        if srcActions==self:
            if position==None:
                return True
            if len(self.keyframes)>0 and position==self.keyframes[0].pos:
                return True
            
        if check:
            if not self.checkAddActionsAt(srcActions, position):
                return False

        actions1 = srcActions.copyAt(position)
        # insert keyframes (after the interval was added, else getInterval()
        # finds these keywords)
        lookup = {}
        for sourcekf in actions1.keyframes:
            destkf = self.addKeyframe(sourcekf)
            lookup[sourcekf] = destkf
            if destkf != sourcekf:
                li = sourcekf.leftInterval
                
                if li:
                    li.kf2 = destkf
                ri = sourcekf.rightInterval
                if ri:
                    ri.kf1 = destkf
        for i, inter in enumerate(actions1.intervals):
            kf1 = actions1.keyframes[i]
            kf2 = actions1.keyframes[i+1]
            if lookup[kf1] != kf1 and lookup[kf2] != kf2:
                # both keyframes overwrote existing keyframes ==> interval needs to be overwritten
                # i.e. we copy the valueGenerator of the srcInterval to the existing interval
                lookup[kf1].rightInterval.valGen = inter.valGen.clone()
            else:
                self._insertInterval(inter)
                inter.kf1.rightInterval = inter
                inter.kf2.leftInterval = inter
        return True


    def addInterval(self, interval):
        """
        add an interval to this Actions objet.
        The interval can be added if:
            1 - both keyframes fall on keyframes AND
              a - these 2 keyframes belong to the same existing interval
                  in this case, the generator of the destination interval is
                  overwritten with the generator from the source interval
              b - or these 2 keyframes belong 2 succesive intervals (i.e fill gap between intervals)

            2 - if both keyframes fall onto nothing, the interval is addded if there is no existing
                 keyframe inside the interval to be added.

            3 - kf1 falls on nothing and kf2 falls on an existing KF
                 we add if there is no existing KF between kf1.pos+1 and kf2.pos-1
                 
            4 - kf2 falls on nothing and kf1 falls on an existing KF
                 we add if there is no existing KF between kf1.pos+1 and kf2.pos-1

            5 - either kf1 or kf2 fall on an existing Interval
               the interval cannot be added
               
        return True if interval was added and False else
        """
        kf1 = interval.kf1
        kf2 = interval.kf2
        
        obj1 = self.getInterval(kf1.pos)
        obj2 = self.getInterval(kf2.pos)

        if isinstance(obj1, Interval) or isinstance(obj2, Interval):
            return False # case 5

        # at this point kf1 and kf2 can either fall on nothing or a KF 

        if self.keyframeBetween(kf1.pos+1, kf2.pos-1):
            return False

        kf1 = self.addKeyframe(kf1)
        kf2 = self.addKeyframe(kf2)

        if obj1 is not None and obj2 is not None: # both fall in KFs
            ri = obj1.rightInterval
            li = obj2.leftInterval
            if ri and ri==li: # case 1.a
                ri.valGen.generator = interval.valGen.generator
                return True

        kf1.rightInterval = interval
        kf2.leftInterval = interval
        interval.kf1 = kf1
        interval.kf2 = kf2
        self._insertInterval(interval)
        return True


    def getInterval(self, position):
        """
        return what is at the position passed as an argument. This can be a keyframe, and interval, or None if there is nothing there
        """
        result = None

        # check if there is a keyframe at this position
        # cannot be done using intervals because of isolated keyframes
        for k in self.keyframes:
            if position == k.pos:
                return k

        # check is position falls within an interval
        for i, inter in enumerate(self.intervals):
            p1 = None
            p2 = None
            kf1 = inter.kf1
            if kf1:
                p1 = kf1.pos 
            else:
                p1 = 0
                
            kf2 = inter.kf2
            if kf2:
                p2 = kf2.pos 
            else:
                p2 = MaxKF
                
            if p1<position and p2>position:
                return inter

        return None

        
    def getValue(self, frame):
        """
        return a value for a given frame, If not KF or Interval exists at this
        position, we return the string 'Nothing There'
        """
        obj = self.getInterval(frame)

        if obj is None:
            return 'Nothing There'

        elif isinstance(obj, KF):
            ri = obj.rightInterval
            li = obj.leftInterval

            # KF--------------AKF
            # ^
            # obj
            # when playing from left to right we need to configure ri.lastValue
            if ri:
                if obj==ri.kf1: # first KF of interval ri
                    ri._lastSetValue = None
                if isinstance(ri.kf2.valueObject, AutoCurrentValueObject):
                    if frame==0:
                        # update interval generators with a value
                        ri.kf2.getValue(value=self.actor().initialValue)
                    else:
                        ri.kf2.getValue() # getting the value forces generator to update lastVal
                    #print 'Akf2 is set to', ri.valGen.generator.firstVal, self.actor().initialValue
                
                elif isinstance(ri.kf2, KFValueFromFunction):
                    ri.kf2.getValue()

                if ri.valGen.generator is not None:
                    if hasattr(ri.valGen.generator, "constant") and  ri.valGen.generator.constant:
                        return ri.getValue(frame)                   
                    
            # KF--------------AKF
            #                  ^
            #                 obj
            # when playing from left to right we need to return the value at this AKF
            # without setting it from the object's current value
            if li:
                if isinstance(obj.valueObject, AutoCurrentValueObject):
                    return li.valGen.generator.lastVal
                
                if  li.valGen.generator is not None:
                    if hasattr(li.valGen.generator, "constant") and  li.valGen.generator.constant:
                        return li.getValue(frame)                

            return obj.getValue()

        else: # we are on an interval
            if obj.active:
                #print 'get frame:', frame, obj.getValue(frame)
                return obj.getValue(frame)

            
       
    def copyAt(self, position):

        if len(self.keyframes)==0:
            return None

        #delta = position - self.keyframes[0].pos
        delta = position
        
        # create Actions object
        na = self.__class__()
        lookup = {}

        # create copies of KFs
        for k in self.keyframes:
            nkf = k.copy(k.pos+delta)
            na.keyframes.append(nkf)
            lookup[k] = nkf

        # create copies of intervals
        for i in self.intervals:
            ni = i.__class__(lookup[i.kf1], lookup[i.kf2], i.valGen.clone())
            if hasattr(i, 'data'):
                ni.data = i.data.copy()
            na.intervals.append(ni)

        return na


    def linkedCopyAt(self, position):

        if len(self.keyframes)==0:
            return None

        #delta = position - self.keyframes[0].pos
        delta = position
        
        # create Actions object
        na = self.__class__()
        lookup = {}

        # create linked copies of all KFs
        for k in self.keyframes:
            nkf = k.linkedCopy(k.pos+delta)
            na.keyframes.append(nkf)
            lookup[k] = nkf

        # create copies of intervals
        pos = self.keyframes[0].pos
        for i in self.intervals:
            ni = i.getLinkedCopy(position)
            ni.kf1 = lookup[i.kf1]
            ni.kf2 = lookup[i.kf2]
            na.intervals.append(ni)

        return na



from Scenario2.keyframes import ValueObject, ValueGenerator,KF, Interval
from Scenario2.interpolators import FloatScalarInterpolator
from Scenario2.actions import Actions
from Scenario2.actor import Actor
from Scenario2.multipleActorsActions import MultipleActorsActions


def test_001MAActionsConstructor():
    """
    test MultipleActorActions object contructor
    """
    # create object with no value
    maa = MultipleActorsActions()
    assert maa.name == 'MAactions'
    assert len(maa.actors) == 0


def dtest_002MAActionsAddActionsAt():
    """
    test adding actions to MultipleActorsActions object
    """
    # create object with no value
    maa = MultipleActorsActions()

    # create an action to be inserted
    a1 = Actions()
    kf1 = KF(0,0)
    kf2 = KF(10,5.)
    a1.addInterval(Interval(kf1, kf2))

    # create an actor
    actor1 = Actor(None, 'actor1')
    
    # add the action for this actor
    maa.AddActions( actor1, a1 )
    assert maa.actions.has_key(actor1)
    assert len(maa.actions) == 1
    #assert maa.actions[actor1][0] == a1
    
    # add a second action over empty space
    #a2 = a1.copyAt( 40 )
    maa.AddActionsAt( actor1, a1, 40)
    assert len(maa.actions[actor1]) == 2
    #assert maa.actions[actor1][0] == a1
    #assert maa.actions[actor1][1] == a2

    # check that adding third actor at 20 will insert in second position
    a3 = a1.copyAt( 20 )
    maa.AddActions( actor1, a3 )
    assert len(maa.actions[actor1]) == 3
    assert maa.actions[actor1][0] == a1
    assert maa.actions[actor1][1] == a3
    assert maa.actions[actor1][2] == a2
    
    # check that adding actor that overlaps fails
    a4 = a1.copyAt( 5 )
    val = maa.AddActions( actor1, a4 )
    assert val==False
    assert len(maa.actions[actor1]) == 3
    assert maa.actions[actor1][0] == a1
    assert maa.actions[actor1][1] == a3
    assert maa.actions[actor1][2] == a2

    a4 = a1.copyAt( 25 )
    val = maa.AddActions( actor1, a4 )
    assert val==False
    assert len(maa.actions[actor1]) == 3
    assert maa.actions[actor1][0] == a1
    assert maa.actions[actor1][1] == a3
    assert maa.actions[actor1][2] == a2
    
